{*******************************************************}
{File:      NCOciDB.PAS                                 }
{Revision:  0.04.00 / 06.11.1999                        }
{Comment:   NC OCI8 VCL: TOCIDatabase, TOCIQuery,       }
{           TOCIStoredProc, ...                         }
{Copyright: (c) 1999, Nica-Com Ltd.                     }
{Author:    Dmitry Arefiev, diman@ncom.ru               }
{*******************************************************}
{$I NCOciDef.inc}
{*$DEFINE DEBUG}

{TODO
    - rewrite StartWait/EndWait code, what must reduce
        CPU utilization and cursor flickering
TODO}

unit NCOciDB;

interface

Uses Windows, SysUtils, Classes, Controls, DB, NCOci, NCOciWrapper, NCOciBuff,
     NCOciParams, NCOciFilter;

type
    TOCIFetchParams = class;
    TOCIDataFormat = class;
    TOCIDatabase = class;
    TOCILongStream = class;
    TOCILOBStream = class;
    TOCIDataSet = class;
    TOCICustomQuery = class;
    TOCIQuery = class;
    TOCIStoredProc = class;

    TOCIFetchParamsValue = (pvPieceBuffSize, pvPieceBuffOwn, pvRowsetSize,
        pvFetchExact, pvFetchAll, pvInrecDataSize);
    TOCIFetchParamsValues = set of TOCIFetchParamsValue;
    TOCIFetchParams = class(TPersistent)
    private
        FAssignedValues: TOCIFetchParamsValues;
        FOwner: TComponent;
        FPieceBuffSize: sb4;
        FPieceBuffOwn: Boolean;
        FRowsetSize: sb4;
        FFetchExact: sb4;
        FInrecDataSize: sb4;
        function GetFetchAll: Boolean;
        procedure SetFetchAll(const Value: Boolean);
        function GetRowsetSize: sb4;
        function GetFetchExact: sb4;
        function GetInrecDataSize: sb4;
        function GetPieceBuffOwn: Boolean;
        function GetPieceBuffSize: sb4;
        procedure SetFetchExact(const Value: sb4);
        procedure SetInrecDataSize(const Value: sb4);
        procedure SetPieceBuffOwn(const Value: Boolean);
        procedure SetPieceBuffSize(const Value: sb4);
        procedure SetRowsetSize(const Value: sb4);
        function GetParentParams: TOCIFetchParams;
        function IsValueOwn(AValue: TOCIFetchParamsValue): Boolean;
        function IsPBFS: Boolean;
        function IsPBOS: Boolean;
        function IsRSS: Boolean;
        function IsFES: Boolean;
        function IsIDSS: Boolean;
    protected
        property ParentParams: TOCIFetchParams read GetParentParams;
    public
        constructor Create(AOwner: TComponent);
        procedure RestoreDefaults;
        procedure Assign(ASource: TPersistent); override;
        procedure TuneVar(AValue: TOCIVariable);
        procedure TuneStmt(AValue: TOCIStatement);
        property AssignedValues: TOCIFetchParamsValues read FAssignedValues;
    published
        // TOCIStatement
        property PieceBuffSize: sb4 read GetPieceBuffSize write SetPieceBuffSize
            stored IsPBFS default OCI_DEF_PIECEBUFFLEN;
        property PieceBuffOwn: Boolean read GetPieceBuffOwn write SetPieceBuffOwn
            stored IsPBOS default False;
        property RowsetSize: sb4 read GetRowsetSize write SetRowsetSize
            stored IsRSS default 10;
        property FetchExact: sb4 read GetFetchExact write SetFetchExact
            stored IsFES default 0;
        property FetchAll: Boolean read GetFetchAll write SetFetchAll
            stored False default False;
        // TOCIVariable
        property InrecDataSize: sb4 read GetInrecDataSize write SetInrecDataSize
            stored IsIDSS default OCI_DEF_INRECDATASIZE;
    end;

    TOCIDataFormatValue = (fvEInteger, fvEBCD, fvELongString, fvEFixedString,
        fvOBoolType, fvOBoolSize, fvOBoolTrue, fvOBoolFalse);
    TOCIDataFormatValues = set of TOCIDataFormatValue;
    TOCIDataFormat = class(TPersistent)
    private
        FOwner: TComponent;
        FAssignedValues: TOCIDataFormatValues;
        FEnableInteger: Boolean;
        FEnableBCD: Boolean;
        FEnableLongString: Boolean;
        FEnableFixedString: Boolean;
        FOBoolType: TOCIVarDataType;
        FOBoolSize: Integer;
        FOBoolTrue: Variant;
        FOBoolFalse: Variant;
        function IsTS: Boolean;
        function IsFS: Boolean;
        function GetEnableBCD: Boolean;
        function GetEnableFixedString: Boolean;
        function GetEnableInteger: Boolean;
        function GetEnableLongString: Boolean;
        function GetOBoolFalse: Variant;
        function GetOBoolSize: Integer;
        function GetOBoolTrue: Variant;
        function GetOBoolType: TOCIVarDataType;
        function GetParentParams: TOCIDataFormat;
        procedure SetEnableBCD(const Value: Boolean);
        procedure SetEnableFixedString(const Value: Boolean);
        procedure SetEnableInteger(const Value: Boolean);
        procedure SetEnableLongString(const Value: Boolean);
        procedure SetOBoolFalse(const Value: Variant);
        procedure SetOBoolSize(const Value: Integer);
        procedure SetOBoolTrue(const Value: Variant);
        procedure SetOBoolType(const Value: TOCIVarDataType);
        function IsValueOwn(AValue: TOCIDataFormatValue): Boolean;
        function IsEIS: Boolean;
        function IsEBCDS: Boolean;
        function IsELSS: Boolean;
        function IsEFSS: Boolean;
        function IsOBTS: Boolean;
        function IsOBSS: Boolean;
    protected
        property ParentParams: TOCIDataFormat read GetParentParams;
    public
        constructor Create(AOwner: TComponent);
        procedure RestoreDefaults;
        procedure Assign(ASource: TPersistent); override;
        function OBool2Value(AValue: Variant): Boolean;
        function Value2OBool(AValue: Boolean): Variant;
        function OBool2ValueSQL(ALeft, ARight: String): String;
        function Value2OBoolSQL(ALeft, ARight: String): String;
        procedure TuneSelItem(AValue: TOCISelectItem);
        property AssignedValues: TOCIDataFormatValues read FAssignedValues;
    published
        // TOCISelectItem
        property EnableInteger: Boolean read GetEnableInteger write SetEnableInteger
            stored IsEIS default False;
        property EnableBCD: Boolean read GetEnableBCD write SetEnableBCD
            stored IsEBCDS default False;
        property EnableLongString: Boolean read GetEnableLongString write SetEnableLongString
            stored IsELSS default False;
        property EnableFixedString: Boolean read GetEnableFixedString write SetEnableFixedString
            stored IsEFSS default False;
        // TOCIParam, TOCIMacro, TOCIStoredProc
        property OBoolType: TOCIVarDataType read GetOBoolType write SetOBoolType
            stored IsOBTS default otInteger;
        property OBoolSize: Integer read GetOBoolSize write SetOBoolSize
            stored IsOBSS default 0;
        property OBoolTrue: Variant read GetOBoolTrue write SetOBoolTrue
            stored IsTS;
        property OBoolFalse: Variant read GetOBoolFalse write SetOBoolFalse
            stored IsFS;
    end;

    TOCILoginField = (lfMode, lfProfile);
    TOCILoginFields = set of TOCILoginField;
    TOCIObjKind = (okService, okDatabase, okSelectable, okPackage, okProc);
    TOCIConnectStringUsage = (suDesignOnly, suLoginUpdate);
    TOCIConnectStringUsages = set of TOCIConnectStringUsage;
{$IFNDEF OCI_D5}
    TLoginEvent = procedure(Sender: TObject; Username, Password: string) of object;
{$ENDIF}

    TOCIDatabase = class({$IFDEF OCI_D5} TCustomConnection {$ELSE}
                                         TOCICustomWrappedComponent {$ENDIF})
    private
{$IFNDEF OCI_D5}
        FDataSets: TList;
        FLoginPrompt: Boolean;
        FStreamedConnected: Boolean;
        FAfterConnect: TNotifyEvent;
        FAfterDisConnect: TNotifyEvent;
        FBeforeConnect: TNotifyEvent;
        FBeforeDisConnect: TNotifyEvent;
        FOnLogin: TLoginEvent;
{$ENDIF}
        FTransaction: TOCITransaction;
        FSession: TOCISession;
        FServer: TOCIServer;
        FService: TOCIService;
        FEnv: TOCIEnvironment;
        FNonBlockingMode: Boolean;
        FDatabaseName: String;
        FServerName: String;
        FUserName: String;
        FPassword: String;
        FAuthentMode: TOCIAuthentMode;
        FInitModes: TOCIDatabaseModes;
        FTransIsolation: TOCITransactionMode;
        FSQLBased: Boolean;
        FInTransaction: Boolean;
        FOnChangePassword: TNotifyEvent;
        FAutoCommit: Boolean;
        FDBUseCount: Integer;
        FSavepoints: TStringList;
        FDefaultFetchParams: TOCIFetchParams;
        FLoginRetries: Integer;
        FLoginFields: TOCILoginFields;
        FInInteractiveOpen: Boolean;
        FWaitCount: Integer;
        FWaitCursor: TCursor;
        FStartTime: Integer;
        FDefaultDataFormat: TOCIDataFormat;
        FAfterCommit: TNotifyEvent;
        FAfterRollback: TNotifyEvent;
        FAfterStartTransaction: TNotifyEvent;
        FBeforeCommit: TNotifyEvent;
        FBeforeRollback: TNotifyEvent;
        FBeforeStartTransaction: TNotifyEvent;
        FSilentMode: Boolean;
        FConnectStringUsages: TOCIConnectStringUsages;
        FLock: TRTLCriticalSection;
        FKeepConnection: Boolean;

        function GetConnectString: String;
        procedure SetAuthentMode(const Value: TOCIAuthentMode);
        procedure SetConnectString(const Value: String);
        procedure SetDatabaseName(const Value: String);
        procedure SetInitModes(const Value: TOCIDatabaseModes);
        procedure SetNonBlockingMode(const Value: Boolean);
        procedure SetPassword(const Value: String);
        procedure SetServerName(const Value: String);
        procedure SetUserName(const Value: String);
{$IFNDEF OCI_D5}
        function GetConnected: Boolean;
        procedure SetConnected(Value: Boolean);
        function GetDataSetCount: Integer;
{$ENDIF}
        function GetDataSet(Index: Integer): TOCIDataSet;
        procedure CheckActive;
        procedure CheckInactive;
        procedure FreeHandles;
        function GetAutoCommit: Boolean;
        procedure SetDefaultFetchParams(AValue: TOCIFetchParams);
        function GetServerVersion: String;
        class procedure GetServicesList(AList: TStrings);
        procedure SetDefaultDataFormat(AValue: TOCIDataFormat);
        procedure DoYield(ASender: TObject);
        function IsDBNS: Boolean;
        procedure SetKeepConnection(Value: Boolean);

    protected
        procedure Loaded; override;
        procedure AttachDataSet(ADS: TOCIDataSet);
        procedure DetachDataSet(ADS: TOCIDataSet);
{$IFDEF OCI_D5}
        procedure DoConnect; override;
        procedure DoDisconnect; override;
        function GetConnected: Boolean; override;
        procedure SetConnected(Value: Boolean); override;
{$ENDIF}

    public
        constructor Create(AOwner: TComponent); override;
        destructor Destroy; override;

{$IFNDEF OCI_D5}
        procedure Close;
        procedure Open;
{$ENDIF}        
        procedure ChangePassword(const ANewPassword: String);
        function InteractiveOpen: Boolean;

        procedure CloseDataSets;
        procedure StartTransaction;
        procedure Commit;
        procedure Rollback;
        procedure RollbackAll;
        procedure CommitAll;
        procedure Break;
        procedure ExecSQL(const ASQL: String);

        procedure StartWait;
        procedure EndWait;

        class function DatabaseByName(const AName: String): TOCIDatabase;
        class procedure GetObjectsList(ADatabase: TOCIDatabase; AList: TStrings;
            const AParentObject: String; AObjKind: TOCIObjKind; AIncludeSystem: Boolean);

{$IFNDEF D5}
        property DataSetCount: Integer read GetDataSetCount;
{$ENDIF}
        property DataSets[Index: Integer]: TOCIDataSet read GetDataSet;

        property IsSQLBased: Boolean read FSQLBased;
        property InTransaction: Boolean read FInTransaction;
        property ServerVersion: String read GetServerVersion;

        property HTransaction: TOCITransaction read FTransaction;
        property HSession: TOCISession read FSession;
        property HServer: TOCIServer read FServer;
        property HService: TOCIService read FService;
        property HEnvironment: TOCIEnvironment read FEnv;

    published
        property DatabaseName: String read FDatabaseName write SetDatabaseName
            stored IsDBNS;

        property NonBlockingMode: Boolean read FNonBlockingMode
            write SetNonBlockingMode default False;
        property InitModes: TOCIDatabaseModes read FInitModes
            write SetInitModes default [];

        property AuthentMode: TOCIAuthentMode read FAuthentMode
            write SetAuthentMode default amDefault;
        property UserName: String read FUserName write SetUserName;
        property Password: String read FPassword write SetPassword;
        property ServerName: String read FServerName write SetServerName;
        property ConnectString: String read GetConnectString
            write SetConnectString stored False;
        property ConnectStringUsages: TOCIConnectStringUsages
            read FConnectStringUsages write FConnectStringUsages
            default [suLoginUpdate];
{$IFNDEF OCI_D5}
        property Connected: Boolean read GetConnected write SetConnected
            default False;
        property LoginPrompt: Boolean read FLoginPrompt write FLoginPrompt
            default True;
{$ELSE}
        property Connected;
        property LoginPrompt default True;
{$ENDIF}
        property LoginRetries: Integer read FLoginRetries write FLoginRetries
            default 3;
        property LoginFields: TOCILoginFields read FLoginFields write FLoginFields
            default [lfMode, lfProfile];
        property KeepConnection: Boolean read FKeepConnection write SetKeepConnection
            default True;

        property AutoCommit: Boolean read GetAutoCommit write FAutoCommit
            default True;
        property TransIsolation: TOCITransactionMode read FTransIsolation
            write FTransIsolation default tmReadWrite;

        property DefaultFetchParams: TOCIFetchParams read FDefaultFetchParams
            write SetDefaultFetchParams;
        property DefaultDataFormat: TOCIDataFormat read FDefaultDataFormat
            write SetDefaultDataFormat;

        property SilentMode: Boolean read FSilentMode write FSilentMode
            default False;
        property WaitCursor: TCursor read FWaitCursor write FWaitCursor
            default crSQLWait;

{$IFNDEF OCI_D5}
        property OnLogin: TLoginEvent read FOnLogin write FOnLogin;
        property BeforeConnect: TNotifyEvent read FBeforeConnect write FBeforeConnect;
        property AfterConnect: TNotifyEvent read FAfterConnect write FAfterConnect;
        property BeforeDisConnect: TNotifyEvent read FBeforeDisConnect write FBeforeDisConnect;
        property AfterDisConnect: TNotifyEvent read FAfterDisConnect write FAfterDisConnect;
{$ELSE}
        property OnLogin;
        property BeforeConnect;
        property AfterConnect;
        property BeforeDisConnect;
        property AfterDisConnect;
{$ENDIF}
        property BeforeStartTransaction: TNotifyEvent read FBeforeStartTransaction write FBeforeStartTransaction;
        property AfterStartTransaction: TNotifyEvent read FAfterStartTransaction write FAfterStartTransaction;
        property BeforeCommit: TNotifyEvent read FBeforeCommit write FBeforeCommit;
        property AfterCommit: TNotifyEvent read FAfterCommit write FAfterCommit;
        property BeforeRollback: TNotifyEvent read FBeforeRollback write FBeforeRollback;
        property AfterRollback: TNotifyEvent read FAfterRollback write FAfterRollback;
        property OnChangePassword: TNotifyEvent read FOnChangePassword write FOnChangePassword;
    end;

    TOCILongStream = class(TMemoryStream)
    private
        FField: TBlobField;
        FMode: TBlobStreamMode;
        FDataSet: TOCIDataSet;
        FRecBuf: PChar;
        FOpened: Boolean;
        FModified: Boolean;
        FParam: TOciParam;
        FIndex: Integer;
        procedure CheckCanModify;
    public
        constructor Create(Field: TBLobField; Mode: TBlobStreamMode);
        constructor CreateUsingParam(Param: TOciParam; AIndex: Integer; Mode: TBlobStreamMode);
        destructor Destroy; override;
        procedure Clear;
        procedure SetSize(NewSize: Longint); override;
        function Write(const Buffer; Count: Longint): Longint; override;
        procedure Truncate;
        procedure Assign(ASource: TObject);
    end;

    TOCILOBStream = class(TStream)
    private
        FOCILobLocator: TOCILobLocator;
        FPosition: sb4;
        FMode: TBlobStreamMode;
        FOpened: Boolean;
        FModified: Boolean;
        FField: TBlobField;
        FDataSet: TOCIDataSet;
        FParam: TOciParam;
        FIndex: Integer;
        procedure CheckOpen;
        procedure CheckWrite;
        constructor Create(Field: TBLobField; Mode: TBlobStreamMode;
            ALobLocatorClass: TOCILobLocatorClass);
        constructor CreateUsingParam(Param: TOciParam; AIndex: Integer;
            Mode: TBlobStreamMode; ALobLocatorClass: TOCILobLocatorClass);
    public
        destructor Destroy; override;
        function IsNull: Boolean;
        function Read(var Buffer; Count: Longint): Longint; override;
        function Write(const Buffer; Count: Longint): Longint; override;
        function Seek(Offset: Longint; Origin: Word): Longint; override;
        procedure Truncate; virtual;
        procedure Assign(ASource: TObject);
    end;

    TOCIFILEStream = class(TOCILOBStream)
    private
        function GetFileName: String;
        procedure SetFileName(const AValue: String);
        function GetDirectory: String;
        procedure SetDirectory(const AValue: String);
        function GetFileExists: Boolean;
        function GetLocator: TOCIExtLocator;
    public
        constructor Create(Field: TBLobField; Mode: TBlobStreamMode);
        constructor CreateUsingParam(Param: TOciParam; AIndex: Integer;
            Mode: TBlobStreamMode);
        property FileExists: Boolean read GetFileExists;
        property FileName: String read GetFileName write SetFileName;
        property Directory: String read GetDirectory write SetDirectory;
        property HLocator: TOCIExtLocator read GetLocator;
    end;

    TOCIILOBStream = class(TOCILOBStream)
    private
        function GetLocator: TOCIIntLocator;
    public
        constructor Create(Field: TBLobField; Mode: TBlobStreamMode);
        constructor CreateUsingParam(Param: TOciParam; AIndex: Integer;
            Mode: TBlobStreamMode);
        function Write(const Buffer; Count: Longint): Longint; override;
        procedure Truncate; override;
        property HLocator: TOCIIntLocator read GetLocator;
    end;

    TOCIDataSet = class(TOCICustomWrappedDataSet)
    private
        FDatabase: TOCIDatabase;
        FDatabaseName: String;
        FFOCICursor: TOCICursor;
        FRecordSize: Word;
        FRecInfoOfs: Word;
        FRecBufSize: Word;
        FDBUseCount: Integer;
        FCachedUpdates: Boolean;
        FFetchParams: TOCIFetchParams;
        FDataFormat: TOCIDataFormat;
        FHaveBLOBs: Boolean;
        FFilters: TOCIFilters;
        FFilterBuffer: PChar;
        FInFindRecord: Boolean;
        FROWIDFieldNo: Integer;
        procedure SetDatabaseName(const Value: String);
        function GetActiveRecBuf(var RecBuf: PChar): Boolean;
        procedure InitBufferPointers;
        function OCIrs2ds(ARS: TOCIRecordStatus): TUpdateStatus;
        procedure SetCachedUpdates(AValue: Boolean);
        procedure SetFetchParams(AValue: TOCIFetchParams);
{$IFDEF DEBUG}
        function GetOCICursor: TOCICursor;
{$ENDIF}
        function GetODataType(AFieldNo: Integer): TOCIVarDataType;
        function GetODataSize(AFieldNo: Integer): ub4;
        procedure SetFilters(AValue: TOCIFilters);
        procedure ExecuteFilter(ASender: TOCICursor; var AAccept: Boolean);
        procedure SetDataFormat(const Value: TOCIDataFormat);
        function IsFS: Boolean;
        function IsDBNS: Boolean;
    protected
        function AllocRecordBuffer: PChar; override;
        procedure FreeRecordBuffer(var Buffer: PChar); override;
        procedure GetBookmarkData(Buffer: PChar; Data: Pointer); override;
        function GetBookmarkFlag(Buffer: PChar): TBookmarkFlag; override;
        function GetRecord(Buffer: PChar; GetMode: TGetMode; DoCheck: Boolean): TGetResult; override;
        function GetRecordSize: Word; override;
        procedure InternalAddRecord(Buffer: Pointer; Append: Boolean); override;
        procedure InternalClose; override;
        procedure InternalDelete; override;
        procedure InternalFirst; override;
        procedure InternalGotoBookmark(Bookmark: Pointer); override;
        procedure InternalHandleException; override;
        procedure InternalInitFieldDefs; override;
        procedure InternalInitRecord(Buffer: PChar); override;
        procedure InternalLast; override;
        procedure InternalOpen; override;
        procedure InternalPost; override;
        procedure InternalSetToRecord(Buffer: PChar); override;
        procedure InternalCancel; override;
        function IsCursorOpen: Boolean; override;
        procedure SetBookmarkFlag(Buffer: PChar; Value: TBookmarkFlag); override;
        procedure SetBookmarkData(Buffer: PChar; Data: Pointer); override;
        procedure SetFieldData(Field: TField; Buffer: Pointer); override;
        procedure InitRecord(Buffer: PChar); override;
        function GetRecordCount: Integer; override;
        function GetRecNo: Integer; override;
        procedure SetRecNo(Value: Integer); override;
        procedure CloseBlob(Field: TField); override;
{$IFNDEF OCI_D5}
        function BCDToCurr(BCD: Pointer; var Curr: Currency): Boolean; override;
        function CurrToBCD(const Curr: Currency; BCD: Pointer; Precision,
            Decimals: Integer): Boolean; override;
{$ENDIF}            
        procedure ClearCalcFields(Buffer: PChar); override;
        function GetCanModify: Boolean; override;
        procedure SetFiltered(Value: Boolean); override;
        procedure SetFilterOptions(Value: TFilterOptions); override;
        procedure SetFilterText(const Value: string); override;
        procedure SetOnFilterRecord(const Value: TFilterRecordEvent); override;
        procedure IncDBUse;
        procedure DecDBUse;
        procedure UpdateFiltering(AFilterChanged: Boolean);
        function FindRecord(Restart, GoForward: Boolean): Boolean; override;
        function LocateRecord(const AKeyFields: string; const AKeyValues: Variant;
            AOptions: TLocateOptions): Boolean;
        procedure OpenCursor(InfoQuery: Boolean); override;
        procedure CloseCursor; override;
        procedure DescribeOCICursor; virtual; abstract;
        procedure CreateOCICursor; virtual; abstract;
        procedure DestroyOCICursor; virtual; abstract;
        function GetSelectItem(APos: ub4): TOCISelectItem; virtual; abstract;
        property FOCICursor: TOCICursor read {$IFDEF DEBUG} GetOCICursor {$ELSE}
            FFOCICursor {$ENDIF} write FFOCICursor;
    public
        constructor Create(AOwner: TComponent); override;
        destructor Destroy; override;
        procedure Disconnect; virtual;
        function GetCurrentRecord(Buffer: PChar): Boolean; override;
        function GetFieldData(Field: TField; Buffer: Pointer): Boolean; overload; override;
        function GetFieldData(FieldNo: Integer; Buffer: Pointer): Boolean; overload; override;
        procedure FetchAll;
        procedure OpenExact(ANRows: sb4);
        function CreateBlobStream(Field: TField; Mode: TBlobStreamMode): TStream; override;
        procedure BlobModified(AField: TField; AIsNull: Boolean);
        function IsSequenced: Boolean; override;
        function CompareBookmarks(Bookmark1, Bookmark2: TBookmark): Integer; override;
        function BookmarkValid(Bookmark: TBookmark): Boolean; override;
        function Lookup(const KeyFields: string; const KeyValues: Variant;
            const ResultFields: string): Variant; override;
        function Locate(const KeyFields: string; const KeyValues: Variant;
            Options: TLocateOptions): Boolean; override;
        property Database: TOCIDatabase read FDatabase;
        property ODataType[AFieldNo: Integer]: TOCIVarDataType read GetODataType;
        property ODataSize[AFieldNo: Integer]: ub4 read GetODataSize;
        property HaveBLOBs: Boolean read FHaveBLOBs;
        property Filters: TOCIFilters read FFilters write SetFilters stored IsFS;
        property Filter stored False;
        property Filtered stored False;
        property FilterOptions stored False;
        property HCursor: TOCICursor read {$IFDEF DEBUG} GetOCICursor {$ELSE}
            FFOCICursor {$ENDIF};
    published
        property Active;
        property AutoCalcFields;
        property BeforeOpen;
        property AfterOpen;
        property BeforeClose;
        property AfterClose;
        property BeforeInsert;
        property AfterInsert;
        property BeforeEdit;
        property AfterEdit;
        property BeforePost;
        property AfterPost;
        property BeforeCancel;
        property AfterCancel;
        property BeforeDelete;
        property AfterDelete;
        property BeforeScroll;
        property AfterScroll;
{$IFDEF OCI_D5}
        property BeforeRefresh;
        property AfterRefresh;
{$ENDIF}
        property OnCalcFields;
        property OnDeleteError;
        property OnEditError;
        property OnNewRecord;
        property OnPostError;
        property DatabaseName: String read FDatabaseName write SetDatabaseName
            stored IsDBNS;
        property CachedUpdates: Boolean read FCachedUpdates write SetCachedUpdates default False;
        property FetchParams: TOCIFetchParams read FFetchParams write SetFetchParams;
        property DataFormat: TOCIDataFormat read FDataFormat write SetDataFormat;
    end;

    TOCIChangingKind = (ckMacros, ckSQL, ckLockParse);
    TOCIChangingKinds = set of TOCIChangingKind;
    TOCIStatementType = (stUnknown, stSelect, stUpdate, stDelete, stInsert,
        stCreate, stDrop, stAlter, stBegin, stDeclare);
    TOCICustomQuery = class(TOCIDataSet)
    private
        FSQL: TStrings;
        FParams: TOciParams;
        FMacros: TOciMacros;
        FStatement: TOCIStatement;
        FPrepared: Boolean;
        FParamCheck, FMacroCheck: Boolean;
        FStreamedPrepared: Boolean;
        FExpandedSQL: String;
        FLocal: Boolean;
        FRowsAffected: Integer;
        FSelfChanging: TOCIChangingKinds;
        FUnidirectional: Boolean;
        FBeforePrepare: TNotifyEvent;
        FAfterPrepare: TNotifyEvent;
        FBeforeUnprepare: TNotifyEvent;
        FAfterUnprepare: TNotifyEvent;
        function ParseSQL(AParams: TOciParams; AMacrosUpd, AMacrosRead: TOciMacros;
            const AQuery: String; ACreateParams, ACreateMacros: Boolean): String;
        procedure DoSQLChange(ASender: TObject);
        procedure SetPrepared(Value: Boolean);
        procedure SetSQL(AValue: TStrings);
        function GetMacroCount: Word;
        function GetParamCount: Word;
        function GetStatementType: TOCIStatementType;
        function GetBindMode: TOCIParamBindMode;
        procedure SetBindMode(const Value: TOCIParamBindMode);
        function IsMS: Boolean;
        function IsPS: Boolean;
    protected
        procedure FreeStatement; virtual;
        procedure CreateStatement; virtual;
        procedure DescribeOCICursor; override;
        procedure CreateOCICursor; override;
        procedure DestroyOCICursor; override;
        procedure InternalRefresh; override;
        function GetSelectItem(APos: ub4): TOCISelectItem; override;
        procedure Loaded; override;
        procedure SetParams(AValue: TOciParams); virtual;
        procedure SetMacros(AValue: TOciMacros); virtual;
        procedure Exec(ATimes: Integer = 0);
        property ParamCheck: Boolean read FParamCheck write FParamCheck default True;
        property MacroCheck: Boolean read FMacroCheck write FMacroCheck default True;
        property SQL: TStrings read FSQL write SetSQL;
        property Text: string read FExpandedSQL;
        property Macros: TOCIMacros read FMacros write SetMacros stored IsMS;
        property MacroCount: Word read GetMacroCount;
        property RowsAffected: Integer read FRowsAffected;
        property StatementType: TOCIStatementType read GetStatementType;
    public
        constructor Create(AOwner: TComponent); override;
        destructor Destroy; override;
        procedure Disconnect; override;
        procedure Prepare;
        procedure Unprepare;
        function ParamByName(const Value: string): TOCIParam;
        function FindParam(const Value: string): TOCIParam;
        function MacroByName(const Value: string): TOCIMacro;
        function FindMacro(const Value: string): TOCIMacro;
        procedure MacrosChanged;
        property ParamCount: Word read GetParamCount;
        property Local: Boolean read FLocal;
        property HStatement: TOCIStatement read FStatement;
    published
        property Prepared: Boolean read FPrepared write SetPrepared default False;
        property Params: TOciParams read FParams write SetParams stored IsPS;
        property Unidirectional: Boolean read FUnidirectional write FUnidirectional
            default False;
        property ParamBindMode: TOCIParamBindMode read GetBindMode write SetBindMode
            default pbByName;
        property BeforePrepare: TNotifyEvent read FBeforePrepare write FBeforePrepare;
        property AfterPrepare: TNotifyEvent read FAfterPrepare write FAfterPrepare;
        property BeforeUnprepare: TNotifyEvent read FBeforeUnprepare write FBeforeUnprepare;
        property AfterUnprepare: TNotifyEvent read FAfterUnprepare write FAfterUnprepare;
        property Filtered;
        property FilterOptions;
        property Filter;
        property Filters;
        property OnFilterRecord;
    end;

    TOCIQuery = class(TOCICustomQuery)
    public
        procedure ExecSQL(ATimes: Integer = 0);
        property Text;
        property RowsAffected;
        property StatementType;
        property MacroCount;
    published
        property ParamCheck;
        property MacroCheck;
        property Macros;
        property SQL;
    end;

    TOCIStoredProc = class(TOCICustomQuery)
    private
        FStoredProcName: String;
        FOPackageName: String;
        FOProcedureName: String;
        FOverload: Word;
        procedure SetStoredProcName(const AValue: String);
        function SplitName(const AProcName: String): Boolean;
        procedure BuildSPQuery(ACreateParams: Boolean);
        procedure SetOPackName(const Value: String);
        procedure SetOProcName(const Value: String);
        procedure SetOverload(const Value: Word);
        procedure ProcNameChanged(AFullName: Boolean);
        procedure ReadParamData(Reader: TReader);
        procedure WriteParamData(Writer: TWriter);
    protected
        procedure CreateStatement; override;
        procedure FreeStatement; override;
        procedure SetParams(AValue: TOciParams); override;
        procedure DefineProperties(Filer: TFiler); override;
    public
        constructor Create(AOwner: TComponent); override;
        procedure ExecProc(ATimes: Integer = 0);
        procedure CopyParams(Value: TParams);
        function DescriptionsAvailable: Boolean;
    published
        property StoredProcName: String read FStoredProcName write SetStoredProcName
            stored False;
        property OPackageName: String read FOPackageName write SetOPackName;
        property OProcedureName: String read FOProcedureName write SetOProcName;
        property Overload: Word read FOverload write SetOverload default 0;
        property Params stored False;
    end;

const
    ot2dt: array [TOCIVarDataType] of TFieldType = (ftUnknown, ftSmallInt,
        ftInteger, ftFloat, ftBCD, ftString, ftFixedChar, ftVarBytes, ftDateTime,
        ftMemo, ftBlob, ftCursor,
{$IFNDEF OCI_D5}
        ftMemo, ftBlob, ftMemo, ftBlob,
{$ELSE}
        ftOraClob, ftOraBlob, ftOraClob, ftOraBlob,
{$ENDIF}
        ftString);
    ot2var: array [TOCIVarDataType] of Integer = (varEmpty, varSmallInt,
        varInteger, varDouble, varCurrency, varString, varString, varString, varDate,
        varString, varString, varEmpty, varEmpty, varEmpty, varEmpty, varEmpty,
        varString);
    dt2ot: array [TFieldType] of TOCIVarDataType = (
        otUnknown, otString, otSmallInt, otInteger, otInteger, otUnknown, otFloat,
        otBCD, otBCD, otDateTime, otDateTime, otDateTime, otRaw, otRaw,
        otInteger, otLongRaw, otLong, otUnknown, otUnknown, otUnknown, otUnknown,
        otUnknown, otCursor, otChar, otUnknown, otInteger, otUnknown, otUnknown,
        otUnknown, otUnknown
{$IFDEF OCI_D5}
        , otBLOB, otCLOB, otUnknown, otUnknown, otUnknown, otUnknown
{$ENDIF}        
        );

    pt2vt: array[TParamType] of TOCIVarType = (
        odUnknown, odIn, odOut, odInOut, odOut);
    vt2pt: array[TOCIVarType] of TParamType = (
        ptUnknown, ptInput, ptOutput, ptInputOutput, ptOutput, ptUnknown);

implementation

Uses Forms, Dialogs, Consts, DBConsts, NCOciMsg, NCOciLoginDlg, NCOciErrorDlg,
     NCOciDM, NCOciUtil, NCOciBreakDlg;

var
    FDatabases: TStringList;

// --------------------------------------------------------------------------
// --------------------------------------------------------------------------
// TOCIFetchParams

constructor TOCIFetchParams.Create(AOwner: TComponent);
begin
    inherited Create;
    FOwner := AOwner;
    RestoreDefaults;
end;

procedure TOCIFetchParams.Assign(ASource: TPersistent);
begin
    if ASource is TOCIFetchParams then begin
        if pvPieceBuffSize in TOCIFetchParams(ASource).FAssignedValues then
            PieceBuffSize := TOCIFetchParams(ASource).PieceBuffSize;
        if pvPieceBuffOwn in TOCIFetchParams(ASource).FAssignedValues then
            PieceBuffOwn := TOCIFetchParams(ASource).PieceBuffOwn;
        if pvInrecDataSize in TOCIFetchParams(ASource).FAssignedValues then
            InrecDataSize := TOCIFetchParams(ASource).InrecDataSize;
        if pvRowsetSize in TOCIFetchParams(ASource).FAssignedValues then
            RowsetSize := TOCIFetchParams(ASource).RowsetSize;
        if pvFetchExact in TOCIFetchParams(ASource).FAssignedValues then
            FetchExact := TOCIFetchParams(ASource).FetchExact;
    end
    else
        inherited Assign(ASource);
end;

procedure TOCIFetchParams.TuneVar(AValue: TOCIVariable);
begin
    AValue.InrecDataSize := InrecDataSize;
end;

procedure TOCIFetchParams.TuneStmt(AValue: TOCIStatement);
begin
    AValue.PieceBuffLen := PieceBuffSize;
    AValue.PieceBuffOwn := PieceBuffOwn;
//    RowsetSize
//    FetchExact
end;

function TOCIFetchParams.GetFetchAll: Boolean;
begin
    Result := FetchExact = -1;
end;

procedure TOCIFetchParams.SetFetchAll(const Value: Boolean);
begin
    if Value then
        FetchExact := -1
    else
        FetchExact := 0;
end;

function TOCIFetchParams.GetRowsetSize: sb4;
begin
    Result := FRowSetSize;
    if (FOwner is TOCIDataSet) and TOCIDataSet(FOwner).FHaveBLOBs then
        Result := 2;
end;

function TOCIFetchParams.IsValueOwn(AValue: TOCIFetchParamsValue): Boolean;
begin
    Result := (FOwner = nil) or (FOwner is TOCIDatabase) or
       (AValue in FAssignedValues) or (TOCIDataSet(FOwner).Database = nil);
end;

function TOCIFetchParams.GetParentParams: TOCIFetchParams;
begin
    Result := (FOwner as TOCIDataSet).Database.DefaultFetchParams;
end;

function TOCIFetchParams.GetFetchExact: sb4;
begin
    if IsValueOwn(pvFetchExact) then
        Result := FFetchExact
    else
        Result := ParentParams.FetchExact;
end;

function TOCIFetchParams.GetInrecDataSize: sb4;
begin
    if IsValueOwn(pvInrecDataSize) then
        Result := FInrecDataSize
    else
        Result := ParentParams.InrecDataSize;
end;

function TOCIFetchParams.GetPieceBuffOwn: Boolean;
begin
    if IsValueOwn(pvPieceBuffOwn) then
        Result := FPieceBuffOwn
    else
        Result := ParentParams.PieceBuffOwn;
end;

function TOCIFetchParams.GetPieceBuffSize: sb4;
begin
    if IsValueOwn(pvPieceBuffSize) then
        Result := FPieceBuffSize
    else
        Result := ParentParams.PieceBuffSize;
end;

procedure TOCIFetchParams.RestoreDefaults;
begin
    FAssignedValues := [];
    FPieceBuffSize := OCI_DEF_PIECEBUFFLEN;
    FPieceBuffOwn := False;
    FInrecDataSize := OCI_DEF_INRECDATASIZE;
    FRowsetSize := IDefRowSetSize;
    FFetchExact := 0;
end;

procedure TOCIFetchParams.SetFetchExact(const Value: sb4);
begin
    if (pvFetchExact in FAssignedValues) and (FFetchExact = Value) then
        Exit;
    FFetchExact := Value;
    Include(FAssignedValues, pvFetchExact);
end;

procedure TOCIFetchParams.SetInrecDataSize(const Value: sb4);
begin
    if (pvInrecDataSize in FAssignedValues) and (FInrecDataSize = Value) then
        Exit;
    FInrecDataSize := Value;
    Include(FAssignedValues, pvInrecDataSize);
end;

procedure TOCIFetchParams.SetPieceBuffOwn(const Value: Boolean);
begin
    if (pvPieceBuffOwn in FAssignedValues) and (FPieceBuffOwn = Value) then
        Exit;
    FPieceBuffOwn := Value;
    Include(FAssignedValues, pvPieceBuffOwn);
end;

procedure TOCIFetchParams.SetPieceBuffSize(const Value: sb4);
begin
    if (pvPieceBuffSize in FAssignedValues) and (FPieceBuffSize = Value) then
        Exit;
    FPieceBuffSize := Value;
    Include(FAssignedValues, pvPieceBuffSize);
end;

procedure TOCIFetchParams.SetRowsetSize(const Value: sb4);
begin
    if (pvRowsetSize in FAssignedValues) and (FRowsetSize = Value) then
        Exit;
    FRowsetSize := Value;
    Include(FAssignedValues, pvRowsetSize);
end;

function TOCIFetchParams.IsPBFS: Boolean;
begin
    Result := pvPieceBuffSize in FAssignedValues;
end;

function TOCIFetchParams.IsPBOS: Boolean;
begin
    Result := pvPieceBuffOwn in FAssignedValues;
end;

function TOCIFetchParams.IsRSS: Boolean;
begin
    Result := pvRowsetSize in FAssignedValues;
end;

function TOCIFetchParams.IsFES: Boolean;
begin
    Result := pvFetchExact in FAssignedValues;
end;

function TOCIFetchParams.IsIDSS: Boolean;
begin
    Result := pvInrecDataSize in FAssignedValues;
end;

// --------------------------------------------------------------------------
// --------------------------------------------------------------------------
// TOCIDataFormat

constructor TOCIDataFormat.Create(AOwner: TComponent);
begin
    inherited Create;
    FOwner := AOwner;
    RestoreDefaults;
end;

procedure TOCIDataFormat.Assign(ASource: TPersistent);
begin
    if ASource is TOCIDataFormat then begin
        if fvOBoolType in TOCIDataFormat(ASource).FAssignedValues then
            OBoolType := TOCIDataFormat(ASource).OBoolType;
        if fvOBoolSize in TOCIDataFormat(ASource).FAssignedValues then
            OBoolSize := TOCIDataFormat(ASource).OBoolSize;
        if fvOBoolTrue in TOCIDataFormat(ASource).FAssignedValues then
            OBoolTrue := TOCIDataFormat(ASource).OBoolTrue;
        if fvOBoolFalse in TOCIDataFormat(ASource).FAssignedValues then
            OBoolFalse := TOCIDataFormat(ASource).OBoolFalse;
        if fvEInteger in TOCIDataFormat(ASource).FAssignedValues then
            EnableInteger := TOCIDataFormat(ASource).EnableInteger;
        if fvEBCD in TOCIDataFormat(ASource).FAssignedValues then
            EnableBCD := TOCIDataFormat(ASource).EnableBCD;
        if fvELongString in TOCIDataFormat(ASource).FAssignedValues then
            EnableLongString := TOCIDataFormat(ASource).EnableLongString;
        if fvEFixedString in TOCIDataFormat(ASource).FAssignedValues then
            EnableFixedString := TOCIDataFormat(ASource).EnableFixedString;
    end
    else
        inherited Assign(ASource);
end;

procedure TOCIDataFormat.RestoreDefaults;
begin
    FAssignedValues := [];
    FOBoolType := otSmallInt;
    FOBoolSize := SizeOf(Integer);
    FOBoolTrue := 1;
    FOBoolFalse := 0;
    FEnableInteger := False;
    FEnableBCD := False;
    FEnableLongString := False;
    FEnableFixedString := False;
end;

function TOCIDataFormat.OBool2Value(AValue: Variant): Boolean;
begin
    try
        Result := not VarIsNull(AValue) and (AValue = FOBoolTrue);
    except
        Result := False;
    end;
end;

function TOCIDataFormat.Value2OBool(AValue: Boolean): Variant;
begin
    if AValue then
        Result := FOBoolTrue
    else
        Result := FOBoolFalse;
end;

function TOCIDataFormat.OBool2ValueSQL(ALeft, ARight: String): String;
begin
    Result := 'if ' + ARight + ' then ' + ALeft + ' := ' + Var2SQL(OBoolTrue, Self) +
        '; else ' + ALeft + ' := ' + Var2SQL(OBoolFalse, Self) + '; end if;';
end;

function TOCIDataFormat.Value2OBoolSQL(ALeft, ARight: String): String;
begin
    Result := ALeft + ' := ' + ARight;
    if VarIsNull(OBoolTrue) or VarIsEmpty(OBoolTrue) then
        Result := Result + ' is null;'
    else
        Result := Result + ' = ' + Var2SQL(OBoolTrue, Self) + ';';
end;

procedure TOCIDataFormat.TuneSelItem(AValue: TOCISelectItem);
begin
    AValue.EnableInteger := EnableInteger;
    AValue.EnableBCD := EnableBCD;
    AValue.EnableLongString := EnableLongString;
    AValue.EnableFixedString := EnableFixedString;
end;

function TOCIDataFormat.GetParentParams: TOCIDataFormat;
begin
    Result := (FOwner as TOCIDataSet).Database.DefaultDataFormat;
end;

function TOCIDataFormat.IsValueOwn(AValue: TOCIDataFormatValue): Boolean;
begin
    Result := (FOwner = nil) or (FOwner is TOCIDatabase) or
       (AValue in FAssignedValues) or (TOCIDataSet(FOwner).Database = nil);
end;

function TOCIDataFormat.GetEnableBCD: Boolean;
begin
    if IsValueOwn(fvEBCD) then
        Result := FEnableBCD
    else
        Result := ParentParams.EnableBCD;
end;

function TOCIDataFormat.GetEnableFixedString: Boolean;
begin
    if IsValueOwn(fvEFixedString) then
        Result := FEnableFixedString
    else
        Result := ParentParams.EnableFixedString;
end;

function TOCIDataFormat.GetEnableInteger: Boolean;
begin
    if IsValueOwn(fvEInteger) then
        Result := FEnableInteger
    else
        Result := ParentParams.EnableInteger;
end;

function TOCIDataFormat.GetEnableLongString: Boolean;
begin
    if IsValueOwn(fvELongString) then
        Result := FEnableLongString
    else
        Result := ParentParams.EnableLongString;
end;

function TOCIDataFormat.GetOBoolFalse: Variant;
begin
    if IsValueOwn(fvOBoolFalse) then
        Result := FOBoolFalse
    else
        Result := ParentParams.OBoolFalse;
end;

function TOCIDataFormat.GetOBoolSize: Integer;
begin
    if IsValueOwn(fvOBoolSize) then
        Result := FOBoolSize
    else
        Result := ParentParams.OBoolSize;
end;

function TOCIDataFormat.GetOBoolTrue: Variant;
begin
    if IsValueOwn(fvOBoolTrue) then
        Result := FOBoolTrue
    else
        Result := ParentParams.OBoolTrue;
end;

function TOCIDataFormat.GetOBoolType: TOCIVarDataType;
begin
    if IsValueOwn(fvOBoolType) then
        Result := FOBoolType
    else
        Result := ParentParams.OBoolType;
end;

procedure TOCIDataFormat.SetEnableBCD(const Value: Boolean);
begin
    if (fvEBCD in FAssignedValues) and (FEnableBCD = Value) then
        Exit;
    FEnableBCD := Value;
    Include(FAssignedValues, fvEBCD);
end;

procedure TOCIDataFormat.SetEnableFixedString(const Value: Boolean);
begin
    if (fvEFixedString in FAssignedValues) and (FEnableFixedString = Value) then
        Exit;
    FEnableFixedString := Value;
    Include(FAssignedValues, fvEFixedString);
end;

procedure TOCIDataFormat.SetEnableInteger(const Value: Boolean);
begin
    if (fvEInteger in FAssignedValues) and (FEnableInteger = Value) then
        Exit;
    FEnableInteger := Value;
    Include(FAssignedValues, fvEInteger);
end;

procedure TOCIDataFormat.SetEnableLongString(const Value: Boolean);
begin
    if (fvELongString in FAssignedValues) and (FEnableLongString = Value) then
        Exit;
    FEnableLongString := Value;
    Include(FAssignedValues, fvELongString);
end;

procedure TOCIDataFormat.SetOBoolFalse(const Value: Variant);
begin
    try
        if (fvOBoolFalse in FAssignedValues) and (FOBoolFalse = Value) then
            Exit;
    except
    end;
    FOBoolFalse := Value;
    Include(FAssignedValues, fvOBoolFalse);
end;

procedure TOCIDataFormat.SetOBoolSize(const Value: Integer);
begin
    if (fvOBoolSize in FAssignedValues) and (FOBoolSize = Value) then
        Exit;
    FOBoolSize := Value;
    Include(FAssignedValues, fvOBoolSize);
end;

procedure TOCIDataFormat.SetOBoolTrue(const Value: Variant);
begin
    try
        if (fvOBoolTrue in FAssignedValues) and (FOBoolTrue = Value) then
            Exit;
    except
    end;
    FOBoolTrue := Value;
    Include(FAssignedValues, fvOBoolTrue);
end;

procedure TOCIDataFormat.SetOBoolType(const Value: TOCIVarDataType);
begin
    if (fvOBoolType in FAssignedValues) and (FOBoolType = Value) then
        Exit;
    FOBoolType := Value;
    Include(FAssignedValues, fvOBoolType);
end;

function TOCIDataFormat.IsEBCDS: Boolean;
begin
    Result := fvEBCD in FAssignedValues;
end;

function TOCIDataFormat.IsEFSS: Boolean;
begin
    Result := fvEFixedString in FAssignedValues;
end;

function TOCIDataFormat.IsEIS: Boolean;
begin
    Result := fvEInteger in FAssignedValues;
end;

function TOCIDataFormat.IsELSS: Boolean;
begin
    Result := fvELongString in FAssignedValues;
end;

function TOCIDataFormat.IsOBSS: Boolean;
begin
    Result := fvOBoolSize in FAssignedValues;
end;

function TOCIDataFormat.IsOBTS: Boolean;
begin
    Result := fvOBoolType in FAssignedValues;
end;

function TOCIDataFormat.IsTS: Boolean;
begin
    Result := fvOBoolTrue in FAssignedValues;
end;

function TOCIDataFormat.IsFS: Boolean;
begin
    Result := fvOBoolFalse in FAssignedValues;
end;

// --------------------------------------------------------------------------
// --------------------------------------------------------------------------
// TOCIDatabase

constructor TOCIDatabase.Create(AOwner: TComponent);
begin
    inherited Create(AOwner);
{$IFNDEF OCI_D5}
    FDataSets := TList.Create;
    FStreamedConnected := False;
{$ENDIF}
    LoginPrompt := True;
    FNonBlockingMode := False;
    FSilentMode := False;
    FAuthentMode := amDefault;
    FInitModes := [];
    FTransIsolation := tmReadWrite;
    FSQLBased := True;
    FSavepoints := TStringList.Create;
    FAutoCommit := True;
    FDefaultFetchParams := TOCIFetchParams.Create(Self);
    FDefaultDataFormat := TOCIDataFormat.Create(Self);
    FLoginRetries := IDefLoginRetries;
    FLoginFields := [lfMode, lfProfile];
    FWaitCursor := crSQLWait;
    FConnectStringUsages := [suLoginUpdate];
    FKeepConnection := True;
    InitializeCriticalSection(FLock);
    try
        DatabaseName := SDefaultDBName
    except
    end;
end;

destructor TOCIDatabase.Destroy;
begin
{$IFDEF OCI_D5}
    inherited Destroy;
{$ELSE}
    Close;
{$ENDIF}
    FSavepoints.Free;
    DatabaseName := '';
{$IFNDEF OCI_D5}
    FDataSets.Free;
{$ENDIF}
    FDefaultFetchParams.Free;
    FDefaultDataFormat.Free;
    DeleteCriticalSection(FLock);
{$IFNDEF OCI_D5}
    inherited Destroy;
{$ENDIF}
end;

function TOCIDatabase.IsDBNS: Boolean;
begin
    Result := FDatabaseName <> SDefaultDBName;
end;

procedure TOCIDatabase.SetDefaultFetchParams(AValue: TOCIFetchParams);
begin
    FDefaultFetchParams.Assign(AValue);
end;

procedure TOCIDatabase.Loaded;
begin
{$IFNDEF OCI_D5}
    inherited Loaded;
{$ENDIF}
    try
        if not (csDesigning in ComponentState) and (suDesignOnly in FConnectStringUsages) then begin
            FServerName := '';
            FUserName := '';
            FPassword := '';
        end;
{$IFNDEF OCI_D5}
        if FStreamedConnected then
            Open;
{$ELSE}
        inherited Loaded;
{$ENDIF}
    except
        if csDesigning in ComponentState then
            Application.HandleException(Self)
        else
            raise;
    end;
end;

function TOCIDatabase.GetDataSet(Index: Integer): TOCIDataSet;
begin
{$IFDEF OCI_D5}
    Result := TOCIDataSet(inherited DataSets[Index]);
{$ELSE}
    Result := TOCIDataSet(FDataSets[Index]);
{$ENDIF}
end;

{$IFNDEF OCI_D5}
function TOCIDatabase.GetDataSetCount: Integer;
begin
    Result := FDataSets.Count;
end;
{$ENDIF}

procedure TOCIDatabase.CloseDataSets;
begin
    EnterCriticalSection(FLock);
    try
        while DataSetCount <> 0 do
            DataSets[DataSetCount - 1].Disconnect;
    finally
        LeaveCriticalSection(FLock);
    end;
end;

procedure TOCIDatabase.AttachDataSet(ADS: TOCIDataSet);
begin
    EnterCriticalSection(FLock);
    try
{$IFDEF OCI_D5}
        RegisterClient(ADS);
{$ELSE}
        FDataSets.Add(ADS);
{$ENDIF}
        if FDBUseCount = 0 then
            Open
        else
            Inc(FDBUseCount);
    finally
        LeaveCriticalSection(FLock);
    end;
end;

procedure TOCIDatabase.DetachDataSet(ADS: TOCIDataSet);
begin
    EnterCriticalSection(FLock);
    try
{$IFDEF OCI_D5}
        UnRegisterClient(ADS);
{$ELSE}
        FDataSets.Remove(ADS);
{$ENDIF}
        if FDBUseCount = 1 then begin
            if not FKeepConnection then
                Close;
        end
        else
            Dec(FDBUseCount);
    finally
        LeaveCriticalSection(FLock);
    end;
end;

procedure TOCIDatabase.FreeHandles;
begin
    FTransaction.Free;
    FTransaction := nil;
    FSession.Free;
    FSession := nil;
    FServer.Free;
    FServer := nil;
    FService.Free;
    FService := nil;
    FEnv.Free;
    FEnv := nil;
end;

function TOCIDatabase.GetConnected: Boolean;
begin
    Result := (FTransaction <> nil);
end;

procedure TOCIDatabase.SetConnected(Value: Boolean);
begin
{$IFNDEF OCI_D5}
    if csReading in ComponentState then
        FStreamedConnected := Value
    else
        if Value then
            Open
        else
            Close;
{$ELSE}
    StartWait;
    EnterCriticalSection(FLock);
    try
        inherited SetConnected(Value);
    finally
        LeaveCriticalSection(FLock);
        EndWait;
    end;
{$ENDIF}
end;

procedure TOCIDatabase.{$IFNDEF OCI_D5} Close {$ELSE} DoDisconnect {$ENDIF};
begin
{$IFNDEF OCI_D5}
    StartWait;
    EnterCriticalSection(FLock);
    try
        if Connected then begin
            if not (csDestroying in ComponentState) and Assigned(FBeforeDisConnect) then
                FBeforeConnect(Self);
{$ENDIF}
            CloseDataSets;
            RollbackAll;
            FreeHandles;
            Dec(FDBUseCount);
{$IFNDEF OCI_D5}
            if not (csDestroying in ComponentState) and Assigned(FAfterDisConnect) then
                FAfterDisConnect(Self);
        end;
    finally
        LeaveCriticalSection(FLock);
        EndWait;
    end;
{$ENDIF}
end;

procedure TOCIDatabase.{$IFNDEF OCI_D5} Open {$ELSE} DoConnect {$ENDIF};
var
    done, logPr: Boolean;
    retry: Integer;
begin
{$IFNDEF OCI_D5}
    StartWait;
    EnterCriticalSection(FLock);
    try
        if not Connected then
        try
            if Assigned(FBeforeConnect) then
                FBeforeConnect(Self);
{$ELSE}
        try
{$ENDIF}
            if not SilentMode then
                TNCOciErrorFrm.HookExceptions;
            FEnv := TOCIEnvironment.Create(dmThreaded in InitModes,
                dmObject in InitModes, dmEvents in InitModes,
                dmShared in InitModes);
            FService := TOCIService.Create(FEnv);
            FService.OnYield := DoYield;
            FServer := TOCIServer.Create(FService);
            FSession := TOCISession.Create(FService);
            done := True;
            logPr := not SilentMode and LoginPrompt and not FInInteractiveOpen;
            retry := 0;
            repeat
                try
                    if logPr then
                        if Assigned(OnLogin) then
                            OnLogin(Self, UserName, Password)
                        else if not TNCOciLoginFrm.Execute(Self, True) then
                            Abort;
                    FServer.Attach(ServerName);
                    try
                        FSession.Start(UserName, Password, AuthentMode);
                    except
                        FServer.Detach;
                        raise;
                    end;
                    done := True;
                except on E: EOCINativeError do
                     // invalid password/user name
                    if (E.Errors[0].ErrorCode = 1017) and not SilentMode and
                       (LoginPrompt or FInInteractiveOpen) then begin
                        Application.HandleException(Self);
                        Inc(retry);
                        if retry = LoginRetries then
                            OCIDBErrorFmt(msgOCIDBLoginRetries, [LoginRetries], Self);
                        done := False;
                        logPr := True;
                    end
                    // password expired
                    else if (E.Errors[0].ErrorCode = 28001) and not SilentMode and
                            (LoginPrompt or FInInteractiveOpen) then begin
                        Application.HandleException(Self);
                        if Assigned(FOnChangePassword) then
                            FOnChangePassword(Self)
                        else if not TNCOciLoginFrm.Execute(Self, False) then
                            Abort;
                        done := False;
                        logPr := False;
                    end
                    else
                        raise;
                end;
            until done;
            FTransaction := TOCITransaction.Create(FService);
            FService.NONBLOCKING_MODE := NonBlockingMode;
            Inc(FDBUseCount);
        except
            FreeHandles;
            raise;
        end;
{$IFNDEF OCI_D5}
        if Assigned(FAfterConnect) then
            FAfterConnect(Self);
    finally
        LeaveCriticalSection(FLock);
        EndWait;
    end;
{$ENDIF}
end;

function TOCIDatabase.InteractiveOpen: Boolean;
begin
    FInInteractiveOpen := True;
    try
        Result := TNCOciLoginFrm.Execute(Self, True);
        if Result then
            Open;
    finally
        FInInteractiveOpen := False;
    end;
end;

procedure TOCIDatabase.SetKeepConnection(Value: Boolean);
begin
    if FKeepConnection <> Value then begin
        FKeepConnection := Value;
        if not Value and (FDBUseCount = 0) then
            Close;
    end;
end;

procedure TOCIDatabase.CheckActive;
begin
    if not Connected then
        OCIDBError(msgOCIDBmbActive, Self);
end;

procedure TOCIDatabase.CheckInactive;
begin
    if Connected then
        if (csDesigning in ComponentState) or FInInteractiveOpen then
            Close
        else
            OCIDBError(msgOCIDBmbInactive, Self);
end;

function TOCIDatabase.GetAutoCommit: Boolean;
begin
    if FInTransaction then
        Result := False
    else
        Result := FAutoCommit;
end;

procedure TOCIDatabase.StartTransaction;
var
    spName: String;
begin
    CheckActive;
    EnterCriticalSection(FLock);
    try
        if Assigned(FBeforeStartTransaction) then
            FBeforeStartTransaction(Self);
        if FInTransaction then begin
            spName := SDefSPPrefix + IntToStr(FSavepoints.Count);
            ExecSQL('SAVEPOINT ' + spName);
            FSavepoints.Add(spName);
        end
        else begin
            FTransaction.Start(TransIsolation);
            FInTransaction := True;
        end;
        if Assigned(FAfterStartTransaction) then
            FAfterStartTransaction(Self);
    finally
        LeaveCriticalSection(FLock);
    end;
end;

procedure TOCIDatabase.Commit;
begin
    CheckActive;
    EnterCriticalSection(FLock);
    try
        if Assigned(FBeforeCommit) then
            FBeforeCommit(Self);
        if FSavepoints.Count > 0 then begin
            FSavepoints.Delete(FSavepoints.Count - 1);
        end
        else begin
            FTransaction.Commit;
            FInTransaction := False;
        end;
        if Assigned(FAfterCommit) then
            FAfterCommit(Self);
    finally
        LeaveCriticalSection(FLock);
    end;
end;

procedure TOCIDatabase.CommitAll;
begin
    while InTransaction do
        Commit;
end;

procedure TOCIDatabase.Rollback;
begin
    CheckActive;
    EnterCriticalSection(FLock);
    try
        if not (csDestroying in ComponentState) and Assigned(FBeforeRollback) then
            FBeforeRollback(Self);
        if FSavepoints.Count > 0 then begin
            ExecSQL('ROLLBACK TO SAVEPOINT ' + FSavepoints[FSavepoints.Count - 1]);
            FSavepoints.Delete(FSavepoints.Count - 1);
        end
        else begin
            FTransaction.Rollback;
            FInTransaction := False;
        end;
        if not (csDestroying in ComponentState) and Assigned(FAfterRollback) then
            FAfterRollback(Self);
    finally
        LeaveCriticalSection(FLock);
    end;
end;

procedure TOCIDatabase.RollbackAll;
begin
    while InTransaction do
        Rollback;
end;

procedure TOCIDatabase.Break;
begin
    if Connected then
    try
        FService.Break;
    finally
        if NonBlockingMode then
            FService.Reset;
    end;
end;

procedure TOCIDatabase.ChangePassword(const ANewPassword: String);
begin
    if FSession = nil then
        CheckActive;
    FSession.ChangePassword(UserName, Password, ANewPassword);
    FPassword := ANewPassword;
end;

function TOCIDatabase.GetConnectString: String;
begin
    if (Password = '') and (UserName = '') then
        Result := '/'
    else begin
        Result := UserName;
        if Password <> '' then
            Result := Result + '/' + Password;
    end;
    if ServerName <> '' then
        Result := Result + '@' + ServerName;
    if AuthentMode <> amDefault then begin
        Result := Result + ' AS ';
        if AuthentMode = amSysDba then
            Result := Result + 'SYSDBA'
        else if AuthentMode = amSysOper then
            Result := Result + 'SYSOPER';
    end;
end;

procedure TOCIDatabase.SetConnectString(const Value: String);
var
    i1, i2, i3: Integer;
    s, Rest: String;
begin
    if ConnectString <> Value then begin
        UserName := '';
        Password := '';
        ServerName := '';
        AuthentMode := amDefault;
        Rest := Value;
        i1 := Pos('/', Rest);
        i2 := Pos('@', Rest);
        i3 := Pos('AS', UpperCase(Rest));
        if (i3 <> 0) and
           ((i3 = 1) or (Rest[i3 - 1] = ' ')) and
           ((i3 + 2 > Length(Rest)) or (Rest[i3 + 2] = ' ')) then begin
            s := UpperCase(Trim(Copy(Rest, i3 + 2, Length(Rest))));
            if s = 'SYSDBA' then
                AuthentMode := amSysDba
            else if s = 'SYSOPER' then
                AuthentMode := amSysOper;
            Rest := Copy(Rest, 1, i3 - 1);
        end;
        if i2 <> 0 then begin
            ServerName := Trim(Copy(Rest, i2 + 1, Length(Rest)));
            Rest := Copy(Rest, 1, i2 - 1);
        end;
        if i1 <> 0 then begin
            Password := Trim(Copy(Rest, i1 + 1, Length(Rest)));
            Rest := Copy(Rest, 1, i1 - 1);
        end;
        UserName := Trim(Copy(Value, 1, Length(Rest)));
    end;
end;

procedure TOCIDatabase.SetAuthentMode(const Value: TOCIAuthentMode);
begin
    if FAuthentMode <> Value then begin
        CheckInactive;
        FAuthentMode := Value;
    end;
end;

procedure TOCIDatabase.SetDatabaseName(const Value: String);
var
    i: Integer;
begin
    EnterCriticalSection(FLock);
    try
        if FDatabaseName <> Value then begin
            CheckInactive;
            FDatabases.AddObject(Value, Self);
            if FDatabases.Find(FDatabaseName, i) then
                FDatabases.Delete(i);
            FDatabaseName := Value;
        end;
    finally
        LeaveCriticalSection(FLock);
    end;
end;

procedure TOCIDatabase.SetInitModes(const Value: TOCIDatabaseModes);
begin
    if FInitModes <> Value then begin
        CheckInactive;
        FInitModes := Value;
    end;
end;

procedure TOCIDatabase.SetNonBlockingMode(const Value: Boolean);
begin
    if FNonBlockingMode <> Value then begin
        FNonBlockingMode := Value;
        if Connected then
            FService.NONBLOCKING_MODE := Value;
    end;
end;

procedure TOCIDatabase.SetPassword(const Value: String);
begin
    if FPassword <> Value then begin
        CheckInactive;
        FPassword := Value;
    end;
end;

procedure TOCIDatabase.SetServerName(const Value: String);
begin
    if FServerName <> Value then begin
        CheckInactive;
        FServerName := Value;
    end;
end;

procedure TOCIDatabase.SetUserName(const Value: String);
begin
    if FUserName <> Value then begin
        CheckInactive;
        FUserName := Value;
    end;
end;

class function TOCIDatabase.DatabaseByName(const AName: String): TOCIDatabase;
var
    i: Integer;
    s: String;
begin
    Result := nil;
    if AName = '' then
        s := SDefaultDBName
    else
        s := AName;
    if FDatabases.Find(s, i) then
        Result := TOCIDatabase(FDatabases.Objects[i])
    else
        OCIDBErrorFmt(msgOCIDBUnknown, [s]);
end;

procedure TOCIDatabase.ExecSQL(const ASQL: String);
begin
    CheckActive;
    with TOCIStatement.Create(FEnv) do
    try
        Prepare(ASQL);
        Execute(FService, 0, False, AutoCommit, False);
    finally
        Free;
    end;
end;

function TOCIDatabase.GetServerVersion: String;
begin
    CheckActive;
    Result := FServer.ServerVersion;
end;

class procedure TOCIDatabase.GetServicesList(AList: TStrings);
var
    InComment, InStr: Boolean;
    pCh, pStParam: PChar;
    s, buff: String;
    BraceLevel: Integer;
    f: TFileStream;
begin
    InitOCI;
    AList.Clear;
    try
        f := TFileStream.Create(FOCITnsNames, fmOpenRead or fmShareDenyWrite);
        try
            SetLength(buff, f.Size);
            f.Read(PChar(Buff)^, f.Size);
        finally
            f.Free;
        end;
    except
        Exit;
    end;
    InComment := False;
    InStr := False;
    BraceLevel := 0;
    pCh := PChar(Buff) - 1;
    repeat
        Inc(pCh);
        case pCh^ of
        '#':
            begin
                if not InComment and not InStr then
                    InComment := True;
            end;
        '''':
            if not InComment then
                InStr := not InStr;
        '(':
            if not InComment and not InStr then
                Inc(BraceLevel);
        ')':
            if not InComment and not InStr then
                Dec(BraceLevel);
        #13, #10:
            if InComment then
                InComment := False;
        'a'..'z', 'A'..'Z':
            if not InComment and not InStr and (BraceLevel = 0) then begin
                pStParam := pCh;
                while pCh^ in ['a'..'z', 'A'..'Z', '#', '$', '_', '.'] do
                    Inc(pCh);
                SetString(s, pStParam, pCh - pStParam);
                AList.Add(s);
                Dec(pCh);
            end;
        end;
    until (pCh^ = #0);
end;

class procedure TOCIDatabase.GetObjectsList(ADatabase: TOCIDatabase; AList: TStrings;
    const AParentObject: String; AObjKind: TOCIObjKind; AIncludeSystem: Boolean);
begin
    AList.Clear;
    case AObjKind of
    okService:
        GetServicesList(AList);
    okDatabase:
        AList.Assign(FDatabases);
    okSelectable:
        if ADatabase <> nil then
            TOCIDM.GetSelectables(ADatabase.DatabaseName, AList, AIncludeSystem);
    okPackage:
        if ADatabase <> nil then
            TOCIDM.GetPackages(ADatabase.DatabaseName, AList, AIncludeSystem);
    okProc:
        if ADatabase <> nil then
            if AParentObject <> '' then
                TOCIDM.GetPackProcs(ADatabase.DatabaseName, AParentObject, AList)
            else
                TOCIDM.GetProcs(ADatabase.DatabaseName, AList, AIncludeSystem);
    end;
end;

procedure TOCIDatabase.SetDefaultDataFormat(AValue: TOCIDataFormat);
begin
    FDefaultDataFormat.Assign(AValue);
end;

procedure TOCIDatabase.EndWait;
begin
    if not SilentMode and (FWaitCount > 0) then begin
        Dec(FWaitCount);
        if (FWaitCount = 0) and (GetCurrentThreadID = MainThreadID) then begin
            Screen.Cursor := crDefault;
            TNCOciBreakFrm.Cancel;
        end;
    end;
end;

procedure TOCIDatabase.StartWait;
begin
    if not SilentMode and (WaitCursor <> crDefault) then begin
        Inc(FWaitCount);
        if (FWaitCount = 1) and (GetCurrentThreadID = MainThreadID) then begin
            FStartTime := GetTickCount;
            Screen.Cursor := WaitCursor;
        end;
    end;
end;

procedure TOCIDatabase.DoYield(ASender: TObject);
const
    SQLDelay: DWORD = 50;
begin
    Application.ProcessMessages;
    if (GetTickCount - FStartTime > SQLDelay) and
       not NCOciBreakFrm.Visible then
        TNCOciBreakFrm.Execute(Self);
end;

// --------------------------------------------------------------------------
// --------------------------------------------------------------------------
// TOCILongStream

procedure AssignError(ASource, ADestination: TObject);
var
    SourceName: string;
begin
    if ASource <> nil then
        SourceName := ASource.ClassName
    else
        SourceName := 'nil';
    raise EConvertError.CreateFmt(SAssignError, [SourceName, ADestination.ClassName]);
end;

constructor TOCILongStream.Create(Field: TBLobField; Mode: TBlobStreamMode);
var
    buff: pUb1;
    sz: sb4;
begin
    inherited Create;
    FField := Field;
    FParam := nil;
    FMode := Mode;
    FDataSet := FField.DataSet as TOCIDataSet;
    if Mode in [bmWrite, bmReadWrite] then begin
        if FField.ReadOnly then
            OCIDBErrorFmt(msgOCIBlobReadOnly, [FField.DisplayName]);
        if not (FDataSet.State in [dsEdit, dsInsert]) then
            OCIDBError(msgDataSetNotEditing);
    end;
    if not (FField.FieldKind in [fkData, fkInternalCalc]) or
       not (FDataSet.ODataType[FField.FieldNo] in [otROWID, otString, otChar, otRaw, otLong, otLongRaw]) or
       not FDataSet.GetActiveRecBuf(FRecBuf) then
        OCIDBErrorFmt(msgOCILongStreamInvalid, [FField.DisplayName]);
    if Mode in [bmRead, bmReadWrite] then begin
        if FDataSet.FOCICursor.GetDataPtr(POCIBookmark(FRecBuf),
           FField.FieldNo, buff, sz) then begin
            Write(buff^, sz);
            Position := 0;
        end;
    end;
    FModified := False;
    FOpened := True;
end;

constructor TOCILongStream.CreateUsingParam(Param: TOciParam; AIndex: Integer; Mode: TBlobStreamMode);
var
    buff: pUb1;
    sz: sb4;
begin
    inherited Create;
    FField := nil;
    FParam := Param;
    FMode := Mode;
    FDataSet := FParam.DataSet as TOCIDataSet;
    FIndex := AIndex;
    if Mode in [bmWrite, bmReadWrite] then begin
        if not (FParam.OParamType in [odIn, odInOut]) then
            OCIDBErrorFmt(msgOCIBlobReadOnly, [FParam.DisplayName]);
    end;
    if not (FParam.ODataType in [otROWID, otString, otChar, otRaw, otLong, otLongRaw]) then
        OCIDBErrorFmt(msgOCILongStreamInvalid, [FParam.DisplayName]);
    if Mode in [bmRead, bmReadWrite] then begin
        if FParam.HVariable.GetDataPtr(AIndex, buff, sz) then begin
            Write(buff^, sz);
            Position := 0;
        end;
    end;
    FModified := False;
    FOpened := True;
end;

destructor TOCILongStream.Destroy;
begin
    if FOpened and FModified then
    try
        if FField <> nil then begin
            FDataSet.FOCICursor.SetData(POCIBookmark(FRecBuf),
                FField.FieldNo, pUb1(Memory), Size);
            TBlobField(FField).Modified := True;
            FDataSet.DataEvent(deFieldChange, Longint(FField));
        end
        else begin
            FParam.HVariable.SetData(FIndex, pUb1(Memory), Size);
        end;
    except
        Application.HandleException(Self);
    end;
    inherited Destroy;
end;

procedure TOCILongStream.CheckCanModify;
begin
    if FOpened and (FMode = bmRead) then
        if FField <> nil then
            OCIDBErrorFmt(msgOCIBlobReadOnly, [FField.DisplayName])
        else
            OCIDBErrorFmt(msgOCIBlobReadOnly, [FParam.DisplayName]);
end;

procedure TOCILongStream.Clear;
begin
    CheckCanModify;
    inherited Clear;
    FModified := True;
end;

procedure TOCILongStream.SetSize(NewSize: Longint);
begin
    CheckCanModify;
    inherited SetSize(NewSize);
    FModified := True;
end;

function TOCILongStream.Write(const Buffer; Count: Longint): Longint;
begin
    CheckCanModify;
    Result := inherited Write(Buffer, Count);
    FModified := True;
end;

procedure TOCILongStream.Truncate;
begin
    SetSize(Position + 1);
end;

procedure TOCILongStream.Assign(ASource: TObject);
begin
    if ASource is TStream then
        CopyFrom(TStream(ASource), 0)
    else
        AssignError(ASource, Self);
end;

// --------------------------------------------------------------------------
// --------------------------------------------------------------------------
// TOCILOBStream

constructor TOCILOBStream.Create(Field: TBLobField; Mode: TBlobStreamMode;
    ALobLocatorClass: TOCILobLocatorClass);
var
    ptr: ppOCIHandle;
    sz: sb4;
    RecBuf: PChar;
begin
    inherited Create;
    FPosition := 0;
    FMode := Mode;
    FOpened := False;
    FModified := False;
    FField := Field;
    FParam := nil;
    FDataSet := FField.DataSet as TOCIDataSet;
    if Mode in [bmWrite, bmReadWrite] then begin
        if FField.ReadOnly then
            OCIDBErrorFmt(msgOCIBlobReadOnly, [FField.DisplayName]);
        if not (FDataSet.State in [dsEdit, dsInsert]) then
            OCIDBError(msgDataSetNotEditing);
    end;
    if not (FField.FieldKind in [fkData, fkInternalCalc]) or
       not (FDataSet.ODataType[FField.FieldNo] in [otCLOB, otBLOB, otCFile, otBFile]) or
       not FDataSet.GetActiveRecBuf(RecBuf) then
        OCIDBErrorFmt(msgOCILOBStreamInvalid, [FField.DisplayName]);
    FDataSet.FOCICursor.GetDataPtr(POCIBookmark(RecBuf), FField.FieldNo, pUb1(ptr), sz);
    FOCILobLocator := ALobLocatorClass.CreateUseHandle(FDataSet.FDatabase.FService, ptr^);
    if (Mode = bmWrite) and not IsNull then
        Truncate;
end;

constructor TOCILOBStream.CreateUsingParam(Param: TOciParam; AIndex: Integer;
    Mode: TBlobStreamMode; ALobLocatorClass: TOCILobLocatorClass);
var
    ptr: ppOCIHandle;
    sz: sb4;
begin
    inherited Create;
    FPosition := 0;
    FMode := Mode;
    FOpened := False;
    FModified := False;
    FField := nil;
    FParam := Param;
    FIndex := AIndex;
    FDataSet := FParam.DataSet as TOCIDataSet;
    if Mode in [bmWrite, bmReadWrite] then begin
        if not (FParam.OParamType in [odIn, odInOut]) then
            OCIDBErrorFmt(msgOCIBlobReadOnly, [FParam.DisplayName]);
    end;
    if not (FParam.ODataType in [otCLOB, otBLOB, otCFile, otBFile]) then
        OCIDBErrorFmt(msgOCILOBStreamInvalid, [FParam.DisplayName]);
    FParam.HVariable.GetDataPtr(AIndex, pUb1(ptr), sz);
    FOCILobLocator := ALobLocatorClass.CreateUseHandle(FDataSet.FDatabase.FService, ptr^);
    if (Mode = bmWrite) and not IsNull then
        Truncate;
end;

destructor TOCILOBStream.Destroy;
begin
    try
        if FModified then
            if FField <> nil then
                FDataSet.BlobModified(FField, IsNull)
            else
                FParam.BlobModified(FIndex, IsNull);
        if FOCILobLocator <> nil then begin
            if FOpened then
                FOCILobLocator.Close;
            FOCILobLocator.Free;
            FOCILobLocator := nil;
        end;
    except
        Application.HandleException(Self);
    end;
    inherited Destroy;
end;

function TOCILOBStream.IsNull: Boolean;
begin
    Result := (FOCILobLocator = nil) or not FOCILobLocator.IsInit;
end;

procedure TOCILOBStream.CheckOpen;
begin
    if not IsNull and not FOpened then begin
        FOCILobLocator.Open(FMode = bmRead);
        FOpened := True;
    end;
end;

procedure TOCILOBStream.CheckWrite;
begin
    if IsNull or (Self is TOCIFILEStream) or (FMode = bmRead) then
        OCIDBError(msgOCIBlobReadOnly);
end;

function TOCILOBStream.Read(var Buffer; Count: Longint): Longint;
begin
    CheckOpen;
    if IsNull then
        Result := 0
    else begin
        Result := Count;
        FOCILobLocator.Read(@Buffer, Count, ub4(Result), FPosition + 1);
        Inc(FPosition, Result);
    end;
end;

function TOCILOBStream.Write(const Buffer; Count: Longint): Longint;
begin
    CheckOpen;
    CheckWrite;
    Result := 0;
end;

function TOCILOBStream.Seek(Offset: Longint; Origin: Word): Longint;
begin
    CheckOpen;
    case Origin of
    soFromBeginning: FPosition := Offset;
    soFromCurrent: Inc(FPosition, Offset);
    soFromEnd:
        if IsNull then
            FPosition := 0
        else
            FPosition := FOCILobLocator.Length + Offset;
    end;
    Result := FPosition;
end;

procedure TOCILOBStream.Truncate;
begin
    CheckOpen;
    CheckWrite;
end;

procedure TOCILOBStream.Assign(ASource: TObject);
begin
    if ASource is TOCILOBStream then begin
        if (FMode = bmRead) then
            OCIDBError(msgOCIBlobReadOnly);
        FOCILobLocator.Assign(TOCILOBStream(ASource).FOCILobLocator);
        FPosition := 0;
        FModified := True;
        CheckOpen;
    end
    else if ASource is TStream then
        CopyFrom(TStream(ASource), 0)
    else
        AssignError(ASource, Self);
end;

constructor TOCIFILEStream.Create(Field: TBLobField; Mode: TBlobStreamMode);
begin
    inherited Create(Field, Mode, TOCIExtLocator);
end;

constructor TOCIFILEStream.CreateUsingParam(Param: TOciParam; AIndex: Integer;
    Mode: TBlobStreamMode);
begin
    inherited CreateUsingParam(Param, AIndex, Mode, TOCIExtLocator);
end;

function TOCIFILEStream.GetFileName: String;
begin
    CheckOpen;
    if FOCILobLocator = nil then
        Result := ''
    else
        Result := TOCIExtLocator(FOCILobLocator).FileName;
end;

procedure TOCIFILEStream.SetFileName(const AValue: String);
begin
    CheckOpen;
    CheckWrite;
    TOCIExtLocator(FOCILobLocator).FileName := AValue;
    FModified := True;
end;

function TOCIFILEStream.GetDirectory: String;
begin
    CheckOpen;
    if FOCILobLocator = nil then
        Result := ''
    else
        Result := TOCIExtLocator(FOCILobLocator).Directory;
end;

procedure TOCIFILEStream.SetDirectory(const AValue: String);
begin
    CheckOpen;
    CheckWrite;
    TOCIExtLocator(FOCILobLocator).Directory := AValue;
    FModified := True;
end;

function TOCIFILEStream.GetFileExists: Boolean;
begin
    CheckOpen;
    if FOCILobLocator = nil then
        Result := False
    else
        Result := TOCIExtLocator(FOCILobLocator).FileExists;
end;

function TOCIFILEStream.GetLocator: TOCIExtLocator;
begin
    Result := TOCIExtLocator(FOCILobLocator);
end;

constructor TOCIILOBStream.Create(Field: TBLobField; Mode: TBlobStreamMode);
begin
    inherited Create(Field, Mode, TOCIIntLocator);
end;

constructor TOCIILOBStream.CreateUsingParam(Param: TOciParam; AIndex: Integer;
    Mode: TBlobStreamMode);
begin
    inherited CreateUsingParam(Param, AIndex, Mode, TOCIIntLocator);
end;

function TOCIILOBStream.Write(const Buffer; Count: Longint): Longint;
begin
    inherited Write(Buffer, Count);
    Result := Count;
    TOCIIntLocator(FOCILobLocator).Write(@Buffer, Count, ub4(Result), FPosition + 1);
    Inc(FPosition, Result);
    FModified := True;
end;

procedure TOCIILOBStream.Truncate;
begin
    inherited Truncate;
    TOCIIntLocator(FOCILobLocator).Trim(FPosition + 1);
    FModified := True;
end;

function TOCIILOBStream.GetLocator: TOCIIntLocator;
begin
    Result := TOCIIntLocator(FOCILobLocator);
end;

// --------------------------------------------------------------------------
// --------------------------------------------------------------------------
// TOCIDataSet

constructor TOCIDataSet.Create(AOwner: TComponent);
begin
    inherited Create(AOwner);
    FFetchParams := TOCIFetchParams.Create(Self);
    FDataFormat := TOCIDataFormat.Create(Self);
    FFilters := TOCIFilters.Create(Self);
end;

destructor TOCIDataSet.Destroy;
begin
    FFetchParams.Free;
    FDataFormat.Free;
    FFilters.Free;
    inherited Destroy;
end;

procedure TOCIDataSet.SetFetchParams(AValue: TOCIFetchParams);
begin
    FFetchParams.Assign(AValue);
end;

procedure TOCIDataSet.SetDataFormat(const Value: TOCIDataFormat);
begin
    FDataFormat.Assign(Value);
end;

{$IFDEF DEBUG}
function TOCIDataSet.GetOCICursor: TOCICursor;
begin
    if FFOCICursor = nil then
        ShowMessage('FFOCICursor = nil');
    Result := FFOCICursor;
end;
{$ENDIF}

procedure TOCIDataSet.OpenCursor(InfoQuery: Boolean);
begin
    IncDBUse;
    try
        if not Active then
            if InfoQuery then
                DescribeOCICursor
            else
                CreateOCICursor;
        inherited OpenCursor(InfoQuery);
    except
        if not IsCursorOpen then
            DecDBUse;
        raise;
    end;
end;

procedure TOCIDataSet.InternalOpen;
begin
    FOCICursor.CachedUpdates := FCachedUpdates;
    FieldDefs.Updated := False;
    FieldDefs.Update;
    if DefaultFields then begin
        CreateFields;
        if FROWIDFieldNo <> -1 then
            with Fields.FieldByNumber(FROWIDFieldNo) do
                ProviderFlags := ProviderFlags + [pfInKey];
    end;
    BindFields(True);
    InitBufferPointers;
    UpdateFiltering(False);
    InternalFirst;
    FOCICursor.OpenComplete;
end;

procedure TOCIDataSet.CloseCursor;
begin
    inherited CloseCursor;
    DestroyOCICursor;
    UpdateFiltering(False);
    DecDBUse;
end;

procedure TOCIDataSet.InternalClose;
begin
    BindFields(False);
    if DefaultFields then
        DestroyFields;
end;

function TOCIDataSet.IsCursorOpen: Boolean;
begin
    Result := (FFOCICursor <> nil) and FOCICursor.Active;
end;

procedure TOCIDataSet.InternalInitFieldDefs;
var
    i, j: Integer;
    item: TOCISelectItem;
    FSize, FPrecision: Integer;
    FType: TFieldType;
    FName: String;
begin
    FieldDefs.Clear;
    i := 1;
    FROWIDFieldNo := -1;
    FHaveBLOBs := False;
    while True do begin
        item := GetSelectItem(i);
        if item = nil then
            Break;
        try
            // Name
            FName := item.NAME;
            j := 0;
            while FieldDefs.IndexOf(FName) >= 0 do begin
                Inc(j);
                FName := item.NAME + '_' + IntToStr(j);
            end;
            // Type & Size
            FSize := 0;
            FPrecision := 0;
            FType := ot2dt[item.DataType];
            case item.DataType of
                otBCD:
                    begin
                        FPrecision := item.PRECISION;
                        FSize := item.DATA_SIZE;
                    end;
                otString, otChar, otRaw, otLong, otLongRaw:
                    FSize := item.DATA_SIZE;
                otCLOB, otBLOB, otCFile, otBFile:
                    begin
                        FSize := item.DATA_SIZE;
                        FHaveBLOBs := True;
                    end;
                otROWID:
                    begin
                        FSize := OCI_DEF_ROWIDLEN;
                        if FROWIDFieldNo = -1 then
                            FROWIDFieldNo := i;
                    end;
            end;
            with FieldDefs.AddFieldDef do begin
                FieldNo := i;
                Name := FName;
                DataType := FType;
                Size := FSize;
                Precision := FPrecision;
                Attributes := [];
                if item.IS_NULL = 0 then
                    Attributes := Attributes + [faRequired];
                if item.DataType in [otCFile, otBFile] then
                    Attributes := Attributes + [faReadOnly];
            end;
        finally
            item.Free;
        end;
        Inc(i);
    end;
end;

procedure TOCIDataSet.InternalHandleException;
begin
    Application.HandleException(Self)
end;

function TOCIDataSet.GetODataType(AFieldNo: Integer): TOCIVarDataType;
var
    item: TOCISelectItem;
begin
    Result := otUnknown;
    item := GetSelectItem(AFieldNo);
    if item <> nil then begin
        Result := item.DataType;
        item.Free;
    end;
end;

function TOCIDataSet.GetODataSize(AFieldNo: Integer): ub4;
var
    item: TOCISelectItem;
begin
    Result := 0;
    item := GetSelectItem(AFieldNo);
    if item <> nil then begin
        Result := item.DataSize;
        item.Free;
    end;
end;

// ------------------------------------------------------------
// Database

procedure TOCIDataSet.Disconnect;
begin
    Close;
end;

procedure TOCIDataSet.IncDBUse;
begin
    if FDBUseCount = 0 then begin
        FDatabase := TOCIDatabase.DatabaseByName(FDatabaseName);
        FDatabase.AttachDataSet(Self);
    end;
    Inc(FDBUseCount);
end;

procedure TOCIDataSet.DecDBUse;
begin
    if FDBUseCount = 1 then begin
        FDatabase.DetachDataSet(Self);
        FDatabase := nil;
    end;
    Dec(FDBUseCount);
end;

procedure TOCIDataSet.SetDatabaseName(const Value: String);
begin
    if csReading in ComponentState then
        FDatabaseName := Value
    else if FDatabaseName <> Value then begin
        CheckInactive;
        if FDatabase <> nil then
            OCIDBError(msgOCIDBmbInactive, Self);
        FDatabaseName := Value;
    end;
end;

// -------------------------------------------------------------
// Buffer related
{---------------------------------------------------------------}
{ Self Bookmark   | Ref Bookmark    | CalcFields | TRecInfo     }
{---------------------------------------------------------------}
{ cursor shadow = | actual record = | .......... | RNo, UpdStat,}
{ = Delphi buff   | = OCI buff      |            | BmkFlag      }
{---------------------------------------------------------------}

type
    PRecInfo = ^TRecInfo;
    TRecInfo = record
        RecordNumber: Longint;
        UpdateStatus: TUpdateStatus;
        BookmarkFlag: TBookmarkFlag;
    end;

procedure TOCIDataSet.InitBufferPointers;
begin
    BookmarkSize := FOCICursor.BookmarkSize;
    FRecordSize := BookmarkSize * 2;
    FRecInfoOfs := FRecordSize + CalcFieldsSize;
    FRecBufSize := FRecInfoOfs + SizeOf(TRecInfo);
end;

function TOCIDataSet.GetRecordSize: Word;
begin
    Result := FRecordSize;
end;

function TOCIDataSet.AllocRecordBuffer: PChar;
begin
    Result := AllocMem(FRecBufSize);
    Move(PChar(FOCICursor.AllocateRecord)^, Result^, BookmarkSize);
end;

procedure TOCIDataSet.FreeRecordBuffer(var Buffer: PChar);
begin
    FOCICursor.FreeRecord(POCIBookmark(Buffer));
    FreeMem(Buffer);
end;

procedure TOCIDataSet.InternalInitRecord(Buffer: PChar);
begin
    FOCICursor.InitRecord(POCIBookmark(Buffer));
    FillChar(Buffer[BookmarkSize], BookmarkSize, 0);
    FillChar(Buffer[RecordSize], FRecBufSize - RecordSize, 0);
end;

procedure TOCIDataSet.ClearCalcFields(Buffer: PChar);
begin
    FillChar(Buffer[RecordSize], CalcFieldsSize, 0);
end;

procedure TOCIDataSet.InitRecord(Buffer: PChar);
begin
    inherited InitRecord(Buffer);
    with PRecInfo(Buffer + FRecInfoOfs)^ do begin
        UpdateStatus := TUpdateStatus(usInserted);
        BookMarkFlag := bfInserted;
        RecordNumber := -1;
    end;
end;

function TOCIDataSet.OCIrs2ds(ARS: TOCIRecordStatus): TUpdateStatus;
const
    rs2ds: array [TOCIRecordStatus] of TUpdateStatus = (usUnmodified, usUnmodified,
        usUnmodified, usInserted, usModified, usDeleted, usUnmodified, usUnmodified,
        usUnmodified, usUnmodified);
begin
    Result := rs2ds[ARS];
end;

function TOCIDataSet.GetRecord(Buffer: PChar; GetMode: TGetMode;
  DoCheck: Boolean): TGetResult;
begin
    Database.StartWait;
    try
        try
            case GetMode of
                gmNext: FOCICursor.Skip(1);
                gmPrior: FOCICursor.Skip(-1);
            end;
            if FOCICursor.EOF then
                Result := grEOF
            else if FOCICursor.BOF then
                Result := grBOF
            else begin
                FOCICursor.GetRecord(POCIBookmark(Buffer), POCIBookmark(Buffer + BookmarkSize));
                with PRecInfo(Buffer + FRecInfoOfs)^ do begin
                    UpdateStatus := OCIrs2ds(FOCICursor.RecordStatus);
                    RecordNumber := FOCICursor.RecNo;
                    BookmarkFlag := bfCurrent;
                end;
                GetCalcFields(Buffer);
                Result := grOK;
            end;
        except
            if DoCheck then
                raise;
            Result := grError;
        end;
    finally
        Database.EndWait;
    end;
end;

function TOCIDataSet.GetCurrentRecord(Buffer: PChar): Boolean;
begin
    if not IsEmpty and (GetBookmarkFlag(ActiveBuffer) = bfCurrent) then begin
        UpdateCursorPos;
        try
            FOCICursor.GetRecord(POCIBookmark(Buffer), POCIBookmark(Buffer + BookmarkSize));
            Result := True;
        except
            Result := False;
        end;
    end
    else
        Result := False;
end;

procedure TOCIDataSet.FetchAll;
begin
    CheckBrowseMode;
    Database.StartWait;
    try
        if not FOCICursor.CursorCanceled then
            FOCICursor.Last;
        CursorPosChanged;
    finally
        Database.EndWait;
    end;
end;

procedure TOCIDataSet.OpenExact(ANRows: sb4);
begin
    CheckInActive;
    FetchParams.FetchExact := ANRows;
    Open;
end;

function TOCIDataSet.GetRecordCount: Integer;
begin
    CheckActive;
    try
        Result := FOCICursor.RecordCount;
    except
        Result := -1;
    end;
end;

function TOCIDataSet.GetRecNo: Integer;
var
    BufPtr: PChar;
begin
    CheckActive;
    if State = dsCalcFields then
        BufPtr := CalcBuffer
    else
        BufPtr := ActiveBuffer;
    Result := PRecInfo(BufPtr + FRecInfoOfs).RecordNumber;
end;

procedure TOCIDataSet.SetRecNo(Value: Integer);
begin
    CheckBrowseMode;
    if Value <> RecNo then begin
        DoBeforeScroll;
        try
            FOCICursor.RecNo := Value;
            Resync([rmCenter]);
            DoAfterScroll;
        except
        end;
    end;
end;

function TOCIDataSet.GetActiveRecBuf(var RecBuf: PChar): Boolean;
begin
    case State of
    dsBlockRead, dsBrowse:
        if IsEmpty then
            RecBuf := nil
        else
            RecBuf := ActiveBuffer;
    dsEdit, dsInsert:
        RecBuf := ActiveBuffer;
    dsSetKey:
        RecBuf := nil {PChar(FKeyBuffer) + SizeOf(TKeyBuffer) ???} ;
    dsCalcFields:
        RecBuf := CalcBuffer;
    dsFilter:
        RecBuf := FFilterBuffer;
    dsNewValue:
        if FOCICursor.InApplyUpdates then
            RecBuf := FOCICursor.GetNewRecord(nil)
        else
            RecBuf := FOCICursor.GetNewRecord(POCIBookmark(ActiveBuffer));
    dsOldValue:
        if FOCICursor.InApplyUpdates then
            RecBuf := FOCICursor.GetOldRecord(nil)
        else
            RecBuf := FOCICursor.GetOldRecord(POCIBookmark(ActiveBuffer));
    else
        RecBuf := nil;
    end;
    Result := RecBuf <> nil;
end;

// -------------------------------------------------------------
// Field related

function TOCIDataSet.GetFieldData(FieldNo: Integer; Buffer: Pointer): Boolean;
var
    RecBuf: PChar;
    sz: sb4;
begin
    try
        Result := GetActiveRecBuf(RecBuf);
        if Result then begin
            Result := FOCICursor.GetData(POCIBookmark(RecBuf), FieldNo, Buffer, sz);
            if Result and (Buffer <> nil) and
               (FieldByNumber(FieldNo).DataType in [ftString, ftFixedChar, ftMemo]) then
                PChar(Buffer)[sz] := #0;
        end;
    except
        if BlockReadSize > 0 then
            Result := False
        else
            raise;
    end;
end;

function TOCIDataSet.GetFieldData(Field: TField; Buffer: Pointer): Boolean;
var
    RecBuf: PChar;
begin
    if Field.FieldNo > 0 then
        Result := GetFieldData(Field.FieldNo, Buffer)
    else begin
        Result := GetActiveRecBuf(RecBuf);
        if Result and (State in [dsBrowse, dsEdit, dsInsert,
                                 dsCalcFields, dsBlockRead]) then begin
            Inc(RecBuf, FRecordSize + Field.Offset);
            Result := Boolean(RecBuf[0]);
            if Result and (Buffer <> nil) then
                Move(RecBuf[1], Buffer^, Field.DataSize);
        end;
    end;
end;

procedure TOCIDataSet.SetFieldData(Field: TField; Buffer: Pointer);
var
    RecBuf: PChar;
begin
    with Field do begin
        if not (State in dsWriteModes) then
            DatabaseError(SNotEditing, Self);
        if (State = dsSetKey) and
           ((FieldNo < 0) {or (FIndexFieldCount > 0) and not IsIndexField ???}) then
            DatabaseErrorFmt(SNotIndexField, [DisplayName]);
        GetActiveRecBuf(RecBuf);
        if FieldNo > 0 then begin
            if State = dsCalcFields then
                DatabaseError(SNotEditing, Self);
            if ReadOnly and not (State in [dsSetKey, dsFilter]) then
                DatabaseErrorFmt(SFieldReadOnly, [DisplayName]);
            Validate(Buffer);
            if FieldKind <> fkInternalCalc then
                FOCICursor.SetData(POCIBookmark(RecBuf), FieldNo, Buffer, -1);
        end
        else begin
            Inc(RecBuf, FRecordSize + Offset);
            Boolean(RecBuf[0]) := LongBool(Buffer);
            if Boolean(RecBuf[0]) then
                Move(Buffer^, RecBuf[1], DataSize);
        end;
        if not (State in [dsCalcFields, dsFilter, dsNewValue]) then
            DataEvent(deFieldChange, Longint(Field));
    end;
end;

function TOCIDataSet.CreateBlobStream(Field: TField; Mode: TBlobStreamMode): TStream;
begin
    Result := nil;
    case ODataType[Field.FieldNo] of
    otROWID, otString, otChar, otRaw, otLong, otLongRaw:
        Result := TOCILongStream.Create(Field as TBLobField, Mode);
    otCLOB, otBLOB:
        Result := TOCIILOBStream.Create(Field as TBLobField, Mode);
    otCFile, otBFile:
        Result := TOCIFILEStream.Create(Field as TBLobField, Mode);
    else
        OCIDBError(msgOCIBlobReadOnly);
    end;
end;

procedure TOCIDataSet.CloseBlob(Field: TField);
begin
    {TODO
        Current LONG/BLOB realization does not need CloseBlob.
        May be later...
    TODO}
end;

procedure TOCIDataSet.BlobModified(AField: TField; AIsNull: Boolean);
var
    RecBuf: PChar;
begin
    GetActiveRecBuf(RecBuf);
    FOCICursor.SetIsNull(POCIBookmark(RecBuf), AField.FieldNo, AIsNull);
    TBlobField(AField).Modified := True;
    DataEvent(deFieldChange, Longint(AField));
end;

{$IFNDEF OCI_D5}
function TOCIDataSet.BCDToCurr(BCD: Pointer; var Curr: Currency): Boolean;
begin
    Curr := PCurrency(BCD)^;
    Result := True;
end;

function TOCIDataSet.CurrToBCD(const Curr: Currency; BCD: Pointer; Precision,
  Decimals: Integer): Boolean;
begin
    PCurrency(BCD)^ := Curr;
    Result := True;
end;
{$ENDIF}

// -------------------------------------------------------------
// Navigation / Editing

procedure TOCIDataSet.InternalFirst;
begin
    Database.StartWait;
    try
        FOCICursor.First;
        FOCICursor.Skip(-1);
    finally
        Database.EndWait;
    end;
end;

procedure TOCIDataSet.InternalLast;
begin
    Database.StartWait;
    try
        FOCICursor.Last;
        FOCICursor.Skip(1);
    finally
        Database.EndWait;
    end;
end;

procedure TOCIDataSet.InternalPost;
begin
    Database.StartWait;
    try
        if State = dsEdit then
            FOCICursor.ModifyRecord(POCIBookmark(ActiveBuffer))
        else
            FOCICursor.InsertRecord(POCIBookmark(ActiveBuffer));
    finally
        Database.EndWait;
    end;
end;

procedure TOCIDataSet.InternalDelete;
begin
    Database.StartWait;
    try
        try
            FOCICursor.DeleteRecord;
        except on E: EOCINativeError do
            // not data found
            if E.Errors[0].ErrorCode <> 100 then
                raise;
        end;
    finally
        Database.EndWait;
    end;
end;

function TOCIDataSet.IsSequenced: Boolean;
begin
    Result := not FFilters.AnyActive;
end;

function TOCIDataSet.GetCanModify: Boolean;
begin
    Result := inherited GetCanModify;
    // ??? Result := FCanModify;
end;

procedure TOCIDataSet.InternalCancel;
begin
    // ???
end;

procedure TOCIDataSet.InternalAddRecord(Buffer: Pointer; Append: Boolean);
begin
    Database.StartWait;
    try
        FOCICursor.InsertRecord(POCIBookmark(Buffer));
    finally
        Database.EndWait;
    end;
end;

// -----------------------------------------------------------
// Bookmark

procedure TOCIDataSet.InternalGotoBookmark(Bookmark: TBookmark);
begin
    FOCICursor.Bookmark := Bookmark;
end;

procedure TOCIDataSet.InternalSetToRecord(Buffer: PChar);
begin
    InternalGotoBookmark((Buffer + BookmarkSize));
end;

function TOCIDataSet.GetBookmarkFlag(Buffer: PChar): TBookmarkFlag;
begin
    Result := PRecInfo(Buffer + FRecInfoOfs).BookmarkFlag;
end;

procedure TOCIDataSet.SetBookmarkFlag(Buffer: PChar; Value: TBookmarkFlag);
begin
    PRecInfo(Buffer + FRecInfoOfs).BookmarkFlag := Value;
end;

procedure TOCIDataSet.GetBookmarkData(Buffer: PChar; Data: Pointer);
begin
    Move((Buffer + BookmarkSize)^, Data^, BookmarkSize);
end;

procedure TOCIDataSet.SetBookmarkData(Buffer: PChar; Data: Pointer);
begin
    Move(Data^, (Buffer + BookmarkSize)^, BookmarkSize);
end;

function TOCIDataSet.CompareBookmarks(Bookmark1, Bookmark2: TBookmark): Integer;
const
    RetCodes: array[Boolean, Boolean] of ShortInt = ((2,-1),(1,0));
begin
    Result := RetCodes[Bookmark1 = nil, Bookmark2 = nil];
    if Result = 2 then
        Result := FOCICursor.CompareBookmarks(Bookmark1, Bookmark2);
end;

function TOCIDataSet.BookmarkValid(Bookmark: TBookmark): Boolean;
begin
    Result := Active;
    if Result then begin
        CursorPosChanged;
        FOCICursor.BookmarkValid(Bookmark);
    end;
end;

// -----------------------------------------------------------
// Cached updates

procedure TOCIDataSet.SetCachedUpdates(AValue: Boolean);
begin
    FCachedUpdates := AValue;
    if Active then
        FOCICursor.CachedUpdates := AValue;
end;


// -----------------------------------------------------------
// Filters, lookup, locate

procedure TOCIDataSet.SetFiltered(Value: Boolean);
begin
    inherited SetFiltered(Value);
    FFilters.DefaultFilter.Active := Value;
end;

procedure TOCIDataSet.SetFilterOptions(Value: TFilterOptions);
begin
    inherited SetFilterOptions(Value);
    FFilters.DefaultFilter.Options := Value;
end;

procedure TOCIDataSet.SetFilterText(const Value: string);
begin
    inherited SetFilterText(Value);
    FFilters.DefaultFilter.Text := Value;
end;

procedure TOCIDataSet.SetOnFilterRecord(const Value: TFilterRecordEvent);
begin
    inherited SetOnFilterRecord(Value);
    FFilters.DefaultFilter.OnFilterRecord := Value;
end;

procedure TOCIDataSet.SetFilters(AValue: TOCIFilters);
begin
    FFilters.Assign(AValue);
end;

procedure TOCIDataSet.ExecuteFilter(ASender: TOCICursor; var AAccept: Boolean);
var
    prevState: TDataSetState;
begin
    {TODO
        If we call some method's, that will access to other
        data in buffer (except Delphi Rec Bookmark) we can
        receive AV (?)
    TODO}
    FFilterBuffer := ASender.Bookmark;
    prevState := SetTempState(dsFilter);
    try
        AAccept := FFilters.Execute;
    finally
        FFilterBuffer := nil;
        RestoreState(prevState);
    end;
end;

procedure TOCIDataSet.UpdateFiltering(AFilterChanged: Boolean);
var
    i: Integer;
    existsFilter: Boolean;
begin
    if csReading in ComponentState then
        Exit;
    existsFilter := False;
    for i := 0 to FFilters.Count - 1 do begin
        if FFilters[i].Active and (FFOCICursor <> nil) then begin
            FFilters[i].Prepare;
            existsFilter := True;
        end
        else if not (FFOCICursor <> nil) then
            FFilters[i].UnPrepare;
    end;
    if existsFilter then
        FOCICursor.OnFilterRecord := ExecuteFilter
    else if FFOCICursor <> nil then
        FOCICursor.OnFilterRecord := nil;
    if not FInFindRecord and AFilterChanged and Active then
        First;
end;

function TOCIDataSet.FindRecord(Restart, GoForward: Boolean): Boolean;
var
    prevFiltered: Boolean;
begin
    CheckBrowseMode;
    Database.StartWait;
    try
        DoBeforeScroll;
        SetFound(False);
        Result := False;
        UpdateCursorPos;
        CursorPosChanged;
        FInFindRecord := True;
        try
            prevFiltered := Filtered;
            try
                Filtered := True;
                if GoForward then begin
                    if Restart then
                        FOCICursor.First
                    else
                        FOCICursor.Skip(1);
                end
                else begin
                    if Restart then
                        FOCICursor.Last
                    else
                        FOCICursor.Skip(-1);
                end;
            finally
                Filtered := prevFiltered;
            end;
        finally
            FInFindRecord := True;
        end;
        if not FOCICursor.EOF and not FOCICursor.BOF then begin
            Resync([rmExact, rmCenter]);
            SetFound(True);
            Result := True;
            DoAfterScroll;
        end;
    finally
        Database.EndWait;
    end;
end;

function TOCIDataSet.LocateRecord(const AKeyFields: string;
  const AKeyValues: Variant; AOptions: TLocateOptions): Boolean;
begin
    CheckBrowseMode;
    Database.StartWait;
    try
        CursorPosChanged;
        FInFindRecord := True;
        try
            with FFilters.LookupFilter do
            try
                PrepareLookupFilter(AKeyFields, AKeyValues, AOptions);
                Active := True;
                FOCICursor.First;
                Result := not FOCICursor.EOF and not FOCICursor.BOF;
                Active := False;
            finally
                UnPrepare;
            end;
        finally
            FInFindRecord := False;
        end;
    finally
        Database.EndWait;
    end;
end;

function TOCIDataSet.Lookup(const KeyFields: string; const KeyValues: Variant;
  const ResultFields: string): Variant;
begin
    Result := Null;
    if LocateRecord(KeyFields, KeyValues, []) then begin
        SetTempState(dsCalcFields);
        try
            CalculateFields(TempBuffer);
            Result := FieldValues[ResultFields];
        finally
            RestoreState(dsBrowse);
        end;
    end;
end;

function TOCIDataSet.Locate(const KeyFields: string; const KeyValues: Variant;
    Options: TLocateOptions): Boolean;
begin
    DoBeforeScroll;
    Result := LocateRecord(KeyFields, KeyValues, Options);
    if Result then begin
        Resync([rmExact, rmCenter]);
        DoAfterScroll;
    end;
end;

function TOCIDataSet.IsFS: Boolean;
begin
    Result := Filters.Count > 0;
end;

function TOCIDataSet.IsDBNS: Boolean;
begin
    Result := FDatabaseName <> SDefaultDBName;
end;

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// TOCICustomQuery

constructor TOCICustomQuery.Create(AOwner: TComponent);
begin
    inherited Create(AOwner);
    FSQL := TStringList.Create;
    TStringList(FSQL).OnChange := DoSQLChange;
    FParams := TOciParams.Create(Self);
    FMacros := TOCIMacros.Create(Self);
    FParamCheck := True;
    FMacroCheck := True;
    FRowsAffected := -1;
    FLocal := False;
end;

destructor TOCICustomQuery.Destroy;
begin
    Destroying;
    Disconnect;
    FSQL.Free;
    FMacros.Free;
    FParams.Free;
    inherited Destroy;
end;

procedure TOCICustomQuery.Loaded;
begin
    try
        if FSQL.Count > 0 then
            DoSQLChange(nil);
        if FStreamedPrepared then
            Prepare;
        inherited Loaded;
    except
        if csDesigning in ComponentState then
            InternalHandleException
        else
            raise;
    end;
end;

procedure TOCICustomQuery.SetSQL(AValue: TStrings);
begin
    FSQL.Assign(AValue);
end;

procedure TOCICustomQuery.Disconnect;
begin
    inherited Disconnect;
    UnPrepare;
end;

function TOCICustomQuery.ParseSQL(AParams: TOciParams; AMacrosUpd, AMacrosRead: TOciMacros;
    const AQuery: String; ACreateParams, ACreateMacros: Boolean): String;
var
    InComment1, InComment2, InStr, InName: Boolean;
    pCh, pStCh, pStParam: PChar;
    s, subst: String;
    i, delta, expLen: Integer;
    IsParam: Boolean;
    m: TOCIMacro;
begin
    Result := AQuery;
    if Result = '' then
        Exit;
    pStCh := PChar(AQuery);
    pCh := pStCh - 1;
    InComment1 := False;
    InComment2 := False;
    InStr := False;
    InName := False;
    delta := 0;
    repeat
        repeat
            Inc(pCh);
        until (pCh^ in ['/', '*', '-', '''', '"', #13, #10, ':', CMacroChar, #0]);
        case pCh^ of
        '/':
            begin
                Inc(pCh);
                if not InComment1 and not InComment2 and not InStr and not InName and
                   (pCh^ = '*') then
                    InComment1 := True
                else
                    Dec(pCh);
            end;
        '*':
            begin
                Inc(pCh);
                if not InComment2 and not InStr and not InName and
                   (pCh^ = '/') then
                    InComment1 := False
                else
                    Dec(pCh);
            end;
        '-':
            begin
                Inc(pCh);
                if not InComment1 and not InStr and not InName and
                   (pCh^ = '-') then
                    InComment2 := True
                else
                    Dec(pCh);
            end;
        '''':
            if not InComment1 and not InComment2 and not InName then
                InStr := not InStr;
        '"':
            if not InComment1 and not InComment2 and not InStr then
                InName := not InName;
        #13, #10:
            begin
                if not InComment1 and InComment2 then
                    InComment2 := False;
                if pCh^ = #13 then
                    Result[pCh - pStCh + 1 + delta] := ' ';
            end;
        ':', CMacroChar:
            if not InComment1 and not InComment2 and not InStr and not InName then begin
                IsParam := pCh^ = ':';
                Inc(pCh);
                if not IsParam or (pCh^ <> '=') then begin
                    pStParam := pCh;
                    while pCh^ in ['0'..'9', 'a'..'z', 'A'..'Z', '#', '$', '_'] do
                        Inc(pCh);
                    SetString(s, pStParam, pCh - pStParam);
                    if IsParam then begin
                        if ACreateParams and (AParams.FindParam(s) = nil) then
                            TOciParam.Create(AParams).Name := s;
                    end
                    else begin
                        m := AMacrosUpd.FindMacro(s);
                        if (m = nil) and ACreateMacros then begin
                            m := TOciMacro.Create(AMacrosUpd);
                            m.Name := s;
                        end;
                        if AMacrosUpd <> AMacrosRead then
                            m := AMacrosRead.FindMacro(s);
                        if m <> nil then
                            subst := m.SQL
                        else
                            subst := '';
                        i := pStParam - pStCh + delta;
                        expLen := Length(subst) - (pCh - pStParam + 1);
                        if expLen < 0 then
                            System.Delete(Result, i, -expLen)
                        else if expLen > 0 then begin
                            SetLength(s, expLen);
                            System.Insert(s, Result, i);
                        end;
                        if subst <> '' then
                            Move(PChar(subst)^, (PChar(Result) + i - 1)^, Length(subst));
                        Inc(delta, expLen);
                    end;
                    Dec(pCh);
                end;
            end;
        end;
    until (pCh^ = #0);
end;

procedure TOCICustomQuery.DoSQLChange(ASender: TObject);
var
    PList: TOciParams;
    MList: TOciMacros;
begin
    if (csReading in ComponentState) then
        Exit;
    if FSelfChanging = [] then begin
        Disconnect;
        if ParamCheck or MacroCheck or (csDesigning in ComponentState) then begin
            PList := nil;
            if ParamCheck then
                PList := TOciParams.Create(Self);
            MList := nil;
            if MacroCheck then
                MList := TOciMacros.Create(Self);
            Include(FSelfChanging, ckLockParse);
            try
                FExpandedSQL := ParseSQL(PList, MList, FMacros, SQL.Text, ParamCheck, MacroCheck);
                if ParamCheck then begin
                    PList.AssignValues(FParams);
                    FParams.Clear;
                    FParams.Assign(PList);
                end;
                if MacroCheck then begin
                    MList.AssignValues(FMacros);
                    FMacros.Clear;
                    FMacros.Assign(MList);
                end;
            finally
                Exclude(FSelfChanging, ckLockParse);
                if ParamCheck then
                    PList.Free;
                if MacroCheck then
                    MList.Free;
            end;
        end
        else
            FExpandedSQL := ParseSQL(FParams, FMacros, FMacros, SQL.Text, False, False);
        DataEvent(dePropertyChange, 0);
    end
    else if not (ckLockParse in FSelfChanging) then begin
        if (ckMacros in FSelfChanging) and not (ckSQL in FSelfChanging) then
            Disconnect;
        FExpandedSQL := ParseSQL(FParams, FMacros, FMacros, SQL.Text, False, False);
    end;
end;

procedure TOCICustomQuery.MacrosChanged;
begin
    try
        Include(FSelfChanging, ckMacros);
        DoSQLChange(nil);
    finally
        Exclude(FSelfChanging, ckMacros);
    end;
end;

procedure TOCICustomQuery.SetPrepared(Value: Boolean);
begin
    CheckInActive;
    if (csReading in ComponentState) then
        FStreamedPrepared := Value
    else if Prepared <> Value then begin
        if Value then begin
            if Assigned(FBeforePrepare) then
                FBeforePrepare(Self);
            IncDBUse;
            try
                CreateStatement;
                FPrepared := True;
            except
                FreeStatement;
                DecDBUse;
                raise;
            end;
            if Assigned(FAfterPrepare) then
                FAfterPrepare(Self);
        end
        else begin
            if Assigned(FBeforeUnPrepare) then
                FBeforeUnPrepare(Self);
            FreeStatement;
            FPrepared := False;
            DecDBUse;
            if Assigned(FAfterUnPrepare) then
                FAfterUnPrepare(Self);
        end;
    end;
end;

procedure TOCICustomQuery.CreateStatement;
var
    i: Integer;
    item: TOCISelectItem;
    pVar: TOCIVariable;
begin
    Database.StartWait;
    try
        FStatement := TOCIStatement.Create(FDatabase.FEnv);
        FFetchParams.TuneStmt(FStatement);
        FStatement.Prepare(FExpandedSQL);
        FParams.Bind(FStatement, True);
        FHaveBLOBs := False;
        if FStatement.STMT_TYPE = OCI_STMT_SELECT then begin
            FOCICursor := TOCIDeadCursor.Create;
            FStatement.Describe(FDatabase.FService);
            i := 1;
            while True do begin
                item := GetSelectItem(i);
                if item = nil then
                    Break;
                try
                    pVar := TOCIVariable.Create(FStatement);
                    FFetchParams.TuneVar(pVar);
                    with pVar do begin
                        Position := i;
                        DataType := item.DataType;
                        DataSize := item.DataSize;
                        VarType := odDefine;
                    end;
                    FHaveBLOBs := FHaveBLOBs or (item.DataType in [otBLOB, otCLOB,
                        otBFILE, otCFILE]);
                    FOCICursor.AddDefVar(pVar);
                finally
                    item.Free;
                end;
                Inc(i);
            end;
        end;
    finally
        Database.EndWait;
    end;
end;

procedure TOCICustomQuery.FreeStatement;
var
    n: Integer;
begin
    try
        FParams.Bind(nil, False);
        if FFOCICursor <> nil then
        try
            n := FOCICursor.DefVarCount;
            while n > 0 do begin
                FOCICursor.DefVar[1].Free;
                FOCICursor.RemoveDefVar(1);
                Dec(n);
            end;
            FOCICursor.Free;
        except
            Application.HandleException(Self);
        end;
        FOCICursor := nil;
        FStatement.Free;
    except
        Application.HandleException(Self);
    end;
    FStatement := nil;
end;

procedure TOCICustomQuery.DescribeOCICursor;
begin
    Prepare;
end;

procedure TOCICustomQuery.CreateOCICursor;
begin
    Database.StartWait;
    try
        Prepare;
        if FFOCICursor = nil then
            OCIDBError(msgOCIStmtCantOpen, Self);
        FParams.Bind(FStatement, True);
        FRowsAffected := -1;
        TOCIDeadCursor(FOCICursor).Open(FStatement, FDatabase.FService,
            FFetchParams.RowsetSize, FFetchParams.FetchExact, FUnidirectional);
    finally
        Database.EndWait;
    end;
end;

procedure TOCICustomQuery.DestroyOCICursor;
begin
    if IsCursorOpen then
        FOCICursor.Close;
    UnPrepare;
end;

function TOCICustomQuery.GetSelectItem(APos: ub4): TOCISelectItem;
begin
    if APos > FStatement.PARAM_COUNT then
        Result := nil
    else begin
        Result := TOCISelectItem.Create(FStatement);
        Result.Position := APos;
        Result.Bind;
        FDataFormat.TuneSelItem(Result);
    end;
end;

procedure TOCICustomQuery.Exec(ATimes: Integer = 0);
begin
    CheckInactive;
    IncDBUse;
    Database.StartWait;
    try
        Prepare;
        if FFOCICursor <> nil then
            OCIDBError(msgOCIStmtCantExec, Self);
        FParams.Bind(FStatement, True);
        FRowsAffected := -1;
        FStatement.Execute(FDatabase.FService, ATimes, False, FDatabase.AutoCommit, False);
        FRowsAffected := FStatement.ROW_COUNT;
    finally
        Database.EndWait;
        DecDBUse;
    end;
end;

procedure TOCICustomQuery.InternalRefresh;
var
    savedKeyValues: Variant;
    locateFields: String;
    i: Integer;
begin
    Database.StartWait;
    try
        locateFields := '';
        for i := 0 to FieldCount - 1 do
        with Fields[i] do
            if not ((DataType in [ftBytes, ftVarBytes]) or IsBlob or
                    (DataSet.Fields[i] is TObjectField) or not (pfInKey in ProviderFlags)) then
                if locateFields <> '' then
                    locateFields := locateFields + ';' + FieldName
                else
                    locateFields := FieldName;
        if locateFields = '' then
            OCIDBError(msgOCIKeyFieldsEmpty, Self);
        savedKeyValues := FieldValues[locateFields];
        try
            DisableControls;
            SetState(dsInactive);
            CloseCursor;
            OpenCursor(False);
            SetState(dsBrowse);
            LocateRecord(locateFields, savedKeyValues, []);
            EnableControls;
        except
            EnableControls;
            SetState(dsInactive);
            CloseCursor;
            raise;
        end;
    finally
        Database.EndWait;
    end;
end;

procedure TOCICustomQuery.Prepare;
begin
    Prepared := True;
end;

procedure TOCICustomQuery.UnPrepare;
begin
    Prepared := False;
end;

function TOCICustomQuery.FindParam(const Value: string): TOCIParam;
begin
    Result := FParams.FindParam(Value);
end;

function TOCICustomQuery.GetParamCount: Word;
begin
    Result := FParams.Count;
end;

function TOCICustomQuery.ParamByName(const Value: string): TOCIParam;
begin
    Result := FParams.ParamByName(Value);
end;

procedure TOCICustomQuery.SetParams(AValue: TOciParams);
begin
    FParams.Assign(AValue);
end;

function TOCICustomQuery.GetBindMode: TOCIParamBindMode;
begin
    Result := FParams.ParamBindMode;
end;

procedure TOCICustomQuery.SetBindMode(const Value: TOCIParamBindMode);
begin
    FParams.ParamBindMode := Value;
end;

function TOCICustomQuery.FindMacro(const Value: string): TOCIMacro;
begin
    Result := FMacros.FindMacro(Value);
end;

function TOCICustomQuery.MacroByName(const Value: string): TOCIMacro;
begin
    Result := FMacros.MacroByName(Value);
end;

function TOCICustomQuery.GetMacroCount: Word;
begin
    Result := FMacros.Count;
end;

procedure TOCICustomQuery.SetMacros(AValue: TOciMacros);
begin
    FMacros.Assign(AValue);
end;

function TOCICustomQuery.IsMS: Boolean;
begin
    Result := Macros.Count > 0;
end;

function TOCICustomQuery.IsPS: Boolean;
begin
    Result := Params.Count > 0;
end;

function TOCICustomQuery.GetStatementType: TOCIStatementType;
const
    st: array[OCI_STMT_SELECT..OCI_STMT_DECLARE] of TOCIStatementType =
        (stSelect, stUpdate, stDelete, stInsert, stCreate,
         stDrop, stAlter, stBegin, stDeclare);
begin
    if Prepared then
        Result := st[FStatement.Stmt_Type]
    else
        Result := stUnknown;
end;

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// TOCIQuery

procedure TOCIQuery.ExecSQL(ATimes: Integer = 0);
begin
    Exec(ATimes);
end;

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// TOCIStoredProc

constructor TOCIStoredProc.Create(AOwner: TComponent);
begin
    inherited Create(AOwner);
    ParamCheck := False;
    MacroCheck := False;
end;

function TOCIStoredProc.SplitName(const AProcName: String): Boolean;
var
    InName: Boolean;
    pCh, pStName: PChar;
    i: Integer;
    obj: array[0..2] of String;
begin
    InName := False;
    pStName := PChar(AProcName);
    pCh := pStName - 1;
    i := 0;
    repeat
        Inc(pCh);
        case pCh^ of
        '"':
            InName := not InName;
        '.':
            if not InName then begin
                if i > 2 then
                    OCIDBError(msgOCIPLSQLNameError, Self);
                SetString(Obj[i], pStName, pCh - pStName);
                Obj[i] := Trim(Obj[i]);
                Inc(i);
                pStName := pCh + 1;
            end;
        '@':
            if not InName then
                OCIDBError(msgOCIPLSQLNotRemote, Self);
        end;
    until (pCh^ = #0);
    if i > 2 then
        OCIDBError(msgOCIPLSQLNameError, Self);
    SetString(Obj[i], pStName, pCh - pStName);
    Obj[i] := Trim(Obj[i]);
    if i = 2 then begin
        FOPackageName := obj[0] + '.' + obj[1];
        FOProcedureName := obj[2];
        Result := False;
    end
    else if i = 1 then begin
        FOPackageName := obj[0];
        FOProcedureName := obj[1];
        Result := True;
    end
    else begin
        FOPackageName := '';
        FOProcedureName := obj[0];
        Result := False;
    end;
end;

procedure TOCIStoredProc.BuildSPQuery(ACreateParams: Boolean);
const
    CRLF: String = #13#10;
var
    sBody, sDecl, sIn, sOut: String;
    descr: TOCIDescribe;
    i, j: Integer;
    objType: ub1;
    numArgs, numProcs: ub2;
    item: TOCISelectItem;

    function CreateParam(const APrefix, AVar: String; AItem: TOCISelectItem;
        AIsResult, AExclFromBody, AIsTable: Boolean): Boolean;
    var
        nm: String;
        vt: TOCIVarType;
        dt: TOCIVarDataType;
        sz: sb4;
        i, k: Integer;
        varTp: TOCIVarType;
        varNm, memberName: String;
        rawTp: ub4;

        function TmpName(const APrefix: String): String;
        begin
            Result := APrefix + 'v' + IntToStr(j);
            Inc(j);
        end;

        function GetItemName(AItem: TOCISelectItem): String;
        begin
            Result := AItem.NAME;
            if (Result = '') and AIsResult and not AIsTable then
                Result := SDefParRESULT;
            Result := APrefix + Result;
        end;

        procedure BodyVar(const AName: String);
        begin
            if not AExclFromBody then
                if AIsResult then
                    sBody := AName + ' := ' + sBody
                else
                    sBody := sBody + AName;
        end;

        procedure AddParam;
        begin
            if ACreateParams then
                with TOciParam(Params.Add) do begin
                    OName := ':' + nm;
                    OParamType := vt;
                    ODataType := dt;
                    if sz <> 0 then
                        ODataSize := sz;
                    IsPLSQLTable := AIsTable;
{$IFDEF DEBUG}
                    ShowMessage(OName + ' PT:' + IntToStr(Integer(OParamType)) +
                        ' DT:' + IntToStr(Integer(ODataType)) + ' DS:' + IntToStr(ODataSize) +
                        ' IsT:' + IntToStr(Integer(IsPLSQLTable)));
{$ENDIF}
                end;
        end;

        procedure CheckTypeSupported(ADataType: ub4);
        begin
            if AIsTable and (ADataType in [_SQLT_REC, _SQLT_BOL]) then
                OCIDBError(msgOCIBadTableType, Self);
        end;

    begin
        FDataFormat.TuneSelItem(AItem);
        nm := GetItemName(AItem);
        vt := AItem.VarType;
        dt := AItem.DataType;
        sz := AItem.DataSize;

        Result := True;
        rawTp := AItem.DATA_TYPE;
        CheckTypeSupported(rawTp);
        case rawTp of
        _SQLT_REC:
            begin
                nm := TmpName(APrefix);
                BodyVar(nm);
                sDecl := sDecl + nm + ' ' + AItem.TYPE_NAME + '.' + AItem.SUB_NAME + ';' + CRLF;
                varTp := AItem.VarType;
                varNm := GetItemName(AItem);
                k := descr.OpenList(OCI_ATTR_LIST_ARGUMENTS);
                for i := 1 to k do begin
                    AItem := descr.SelectItem[i];
                    memberName := nm + '.' + AItem.name;
                    try
                        if CreateParam(varNm + SDefParMembDelim, memberName, AItem, AIsResult, True, False) then begin
                            if varTp in [odIn, odInOut] then
                                sIn := sIn + memberName + ' := :' + varNm + SDefParMembDelim + AItem.name + ';' + CRLF;
                            if varTp in [odOut, odInOut] then
                                sOut := sOut + ':' + varNm + SDefParMembDelim + AItem.name + ' := ' + memberName + ';' + CRLF;
                        end;
                    finally
                        AItem.Free;
                    end;
                end;
                descr.CloseList;
                Result := False;
            end;
        _SQLT_TAB:
            begin
                descr.OpenList(OCI_ATTR_LIST_ARGUMENTS);
                AItem := descr.SelectItem[1];
                try
                    Result := CreateParam(nm, AVar, AItem, AIsResult, AExclFromBody, True);
                finally
                    AItem.Free;
                end;
                descr.CloseList;
            end;
        _SQLT_BOL:
            begin
                if AVar = '' then begin
                    varNm := TmpName(APrefix);
                    sDecl := sDecl + varNm + ' boolean;' + CRLF;
                end
                else
                    varNm := AVar;
                if AItem.VarType in [odIn, odInOut] then
                    sIn := sIn + DataFormat.Value2OBoolSQL(varNm, ':' + nm) + CRLF;
                if AItem.VarType in [odOut, odInOut] then
                    sOut := sOut + DataFormat.Value2OBoolSQL(':' + nm, varNm) + CRLF;
                dt := DataFormat.OBoolType;
                sz := DataFormat.OBoolSize;
                BodyVar(varNm);
                AddParam;
                Result := False;
            end;
        else
            BodyVar(':' + nm);
            AddParam;
        end;
    end;

begin
    if FStoredProcName = '' then
        OCIDBError(msgOCIPLSQLObjNameEmpty, Self);
    Database.StartWait;
    try
        descr := TOCIDescribe.Create(FDatabase.FService);
        SQL.BeginUpdate;
        try
            // Describe object and verify object type ...
            if FOPackageName <> '' then
                descr.DescribeName(FOPackageName)
            else
                descr.DescribeName(FOProcedureName);
            objType := descr.UB1[OCI_ATTR_PTYPE];
            if not (objType in [OCI_PTYPE_PKG, OCI_PTYPE_FUNC, OCI_PTYPE_PROC]) or
               (objType = OCI_PTYPE_PKG) and (FOProcedureName = '') or
               (objType = OCI_PTYPE_PKG) and (FOPackageName = '') then
                OCIDBErrorFmt(msgOCINotPLSQLObj, [FStoredProcName], Self);
            // If object is package then locate procedure in him
            if objType = OCI_PTYPE_PKG then begin
                numProcs := descr.OpenList(OCI_ATTR_LIST_SUBPROGRAMS);
                i := 0;
                while i < numProcs do begin
                    descr.GoToItem(i);
                    if (descr.Text[OCI_ATTR_NAME] = FOProcedureName) and  // ??? may be "name"
                       (descr.UB2[OCI_ATTR_OVERLOAD_ID] = FOverload) then
                        Break;
                    Inc(i);
                end;
                if i = numProcs then
                    OCIDBErrorFmt(msgOCINotPackageProc, [FOProcedureName, FOPackageName], Self);
                objType := descr.UB1[OCI_ATTR_PTYPE];
            end;
            // Now start form a SQL and Params
            sDecl := '';
            sIn := '';
            sOut := '';
            sBody := '';
            j := 0;
            numArgs := descr.OpenList(OCI_ATTR_LIST_ARGUMENTS);
            // If this is function then form :Result parameter
            if objType = OCI_PTYPE_FUNC then begin
                item := descr.SelectItem[0];
                try
                    CreateParam('', '', item, True, False, False);
                finally
                    item.Free;
                end;
                Dec(numArgs);
            end;
            // Now for each parameter do ...
            sBody := sBody + FStoredProcName;
            if numArgs > 0 then
                sBody := sBody + '(';
            for i := 1 to numargs do begin
                if i > 1 then
                    sBody := sBody + ', ';
                item := descr.SelectItem[i];
                try
                    CreateParam('', '', item, False, False, False);
                finally
                    item.Free;
                end;
            end;
            if numArgs > 0 then
                sBody := sBody + ')';
            // Time to form anonymous PL/SQL block
            if sDecl <> '' then begin
                SQL.Add('declare');
                SQL.Add(sDecl);
            end;
            SQL.Add('begin');
            if sIn <> '' then
                SQL.Add(sIn);
            SQL.Add(sBody + ';');
            if sOut <> '' then
                SQL.Add(sOut);
            SQL.Add('end;');
            descr.CloseList;
            if FOPackageName <> '' then
                descr.CloseList;
{$IFDEF DEBUG}
            ShowMessage(SQL.Text);
{$ENDIF}
        finally
            descr.Free;
            SQL.EndUpdate;
        end;
    finally
        Database.EndWait;
    end;
end;

procedure TOCIStoredProc.CreateStatement;
begin
    Include(FSelfChanging, ckSQL);
    try
        try
            BuildSPQuery(Params.Count = 0);
            inherited CreateStatement;
        except
            Params.Clear;
            SQL.Clear;
            raise;
        end;
    finally
        Exclude(FSelfChanging, ckSQL);
    end;
end;

procedure TOCIStoredProc.FreeStatement;
begin
    Include(FSelfChanging, ckSQL);
    try
        try
            inherited FreeStatement;
        finally
            Params.Clear;
            SQL.Clear;
        end;
    finally
        Exclude(FSelfChanging, ckSQL);
    end;
end;

procedure TOCIStoredProc.ExecProc(ATimes: Integer = 0);
begin
    Exec(ATimes);
end;

procedure TOCIStoredProc.ProcNameChanged(AFullName: Boolean);
begin
    if not AFullName and (FOProcedureName <> '') then
        if FOPackageName <> '' then
            FStoredProcName := FOPackageName + '.' + FOProcedureName
        else
            FStoredProcName := FOProcedureName;
    if not (csReading in ComponentState) then begin
        SQL.Clear;
        Params.Clear;
        if (csDesigning in ComponentState) and (FStoredProcName <> '') then
            Prepare;
    end;
end;

procedure TOCIStoredProc.SetStoredProcName(const AValue: String);
var
    done, ambiguousName: Boolean;
begin
    if FStoredProcName <> AValue then
        if not (csReading in ComponentState) then begin
            CheckInactive;
            if AValue <> '' then
                ambiguousName := SplitName(AValue)
            else begin
                ambiguousName := False;
                FOPackageName := '';
                FOProcedureName := '';
            end;
            FStoredProcName := AValue;
            done := True;
            repeat
                try
                    ProcNameChanged(True);
                    done := True;
                except on E: EOCINativeError do
                     // object does not exists && name may be standalone proc/func
                    if (E.Errors[0].ErrorCode = 4043) and ambiguousName then begin
                        done := False;
                        ambiguousName := False;
                        FOProcedureName := FOPackageName + '.' + FOProcedureName;
                        FOPackageName := '';
                    end
                    else
                        raise;
                end;
            until done;
        end
        else begin
            FStoredProcName := AValue;
            ProcNameChanged(True);
        end;
end;

procedure TOCIStoredProc.SetOPackName(const Value: String);
begin
    if FOPackageName <> Value then begin
        if not (csReading in ComponentState) then begin
            CheckInactive;
            FStoredProcName := '';
            FOPackageName := Value;
        end
        else
            FOPackageName := Value;
        ProcNameChanged(False);
    end;
end;

procedure TOCIStoredProc.SetOProcName(const Value: String);
begin
    if FOProcedureName <> Value then begin
        if not (csReading in ComponentState) then begin
            CheckInactive;
            FStoredProcName := '';
            FOProcedureName := Value;
        end
        else
            FOProcedureName := Value;
        ProcNameChanged(False);
    end;
end;

procedure TOCIStoredProc.SetOverload(const Value: Word);
begin
    if FOverload <> Value then begin
        if not (csReading in ComponentState) then begin
            CheckInactive;
            FOverload := Value;
        end
        else
            FOverload := Value;
        ProcNameChanged(True);
    end;
end;

procedure TOCIStoredProc.SetParams(AValue: TOciParams);
begin
    if not Prepared and (FParams.Count = 0) then
        try
            Prepare;
            inherited SetParams(AValue);
        finally
            UnPrepare;
        end
    else
        inherited SetParams(AValue);
end;

function TOCIStoredProc.DescriptionsAvailable: Boolean;
var
    prevPrepare: Boolean;
begin
    prevPrepare := Prepared;
    try
        Prepared := True;
        Result := True;
    except
        Result := False;
    end;
    Prepared := prevPrepare;
end;

procedure TOCIStoredProc.CopyParams(Value: TParams);
begin
    if not Prepared and (FParams.Count = 0) then
        try
            Prepare;
            Value.Assign(FParams);
        finally
            UnPrepare;
        end
    else
        Value.Assign(FParams);
end;

procedure TOCIStoredProc.DefineProperties(Filer: TFiler);

    function WriteData: Boolean;
    begin
        if Filer.Ancestor <> nil then
            Result := not FParams.IsEqual(TOCIStoredProc(Filer.Ancestor).FParams)
        else
            Result := FParams.Count > 0;
    end;

begin
    inherited DefineProperties(Filer);
    Filer.DefineProperty('ParamData', ReadParamData, WriteParamData, WriteData);
end;

procedure TOCIStoredProc.ReadParamData(Reader: TReader);
begin
    Reader.ReadValue;
    Reader.ReadCollection(Params);
end;

procedure TOCIStoredProc.WriteParamData(Writer: TWriter);
begin
    Writer.WriteCollection(Params);
end;

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

initialization

    FDatabases := TStringList.Create;
    FDatabases.Sorted := True;
    FDatabases.Duplicates := dupError;

finalization

    FDatabases.Free;
    FDatabases := nil;

end.

